"""
Classes used to provide grouping leaves mechanism.

This file is a part of the program kupfer, which is
released under GNU General Public License v3 (or any later version),
see the main program file, and COPYING for details.
"""

from __future__ import annotations

import copy
import itertools
import time
import typing as ty
import weakref
from collections import defaultdict

from gi.repository import Gdk, Gtk

from kupfer.obj.base import Action, Leaf, Source
from kupfer.support import itertools as kitertools
from kupfer.support import kupferstring

if ty.TYPE_CHECKING:
    from gettext import gettext as _

    from kupfer.core import commandexec

__author__ = (
    "Karol Będkowski <karol.bedkowsk+gh@gmail.com>, "
    "Ulrik Sverdrup <ulrik.sverdrup@gmail.com>"
)


__all__ = ["GroupingLeaf", "GroupingSource", "ToplevelGroupingSource"]

Slots = ty.Optional[dict[str, ty.Any]]


class GroupingLeaf(Leaf):
    """A Leaf that groups with other leaves inside Grouping Sources.

    The represented object of a GroupedLeaf is a dictionary of (slot, value)
    pairs, where slot identifies the slot, and the value is something that must
    be equal to be grouped.

    The GroupingLeaf must have a value for all @grouping_slots, but values of
    None will not be grouped with others.
    """

    grouping_slots: tuple[str, ...] = ()

    def __init__(self, obj: ty.Any, name: str) -> None:
        Leaf.__init__(self, obj, name)
        self.links = [self]

    def slots(self) -> Slots:
        return ty.cast("Slots", self.object)

    def has_content(self) -> bool:
        return len(self.links) > 1

    def content_source(self, alternate: bool = False) -> Source:
        return _GroupedItemsSource(self)

    def get(self, key: ty.Any, default: ty.Any = None) -> ty.Any:
        try:
            return next(self.all(key))
        except StopIteration:
            return default

    def __len__(self) -> int:
        return len(self.links)

    def __contains__(self, key: ty.Any) -> bool:
        "Return True if GroupedLeaf has value for @key"
        return any(key in leaf.object for leaf in self.links)

    def __getitem__(self, key: ty.Any) -> ty.Any:
        "Get first (canonical) value for key"
        try:
            return next(self.all(key))
        except StopIteration as exc:
            raise KeyError(f"{self} has no slot {key}") from exc

    def all(self, key: ty.Any) -> ty.Iterator[ty.Any]:
        "Return iterator of all values for @key"
        return (leaf.object[key] for leaf in self.links if key in leaf.object)

    def check_key(self, key: ty.Any) -> bool:
        """check if GroupedLeaf has non empty value for @key"""
        return any(bool(leaf.object.get(key)) for leaf in self.links)

    def get_actions(self) -> ty.Iterable[Action]:
        for key, val in kitertools.unique_iterator(
            keyval for leaf in self.links for keyval in leaf.object.items()
        ):
            yield CopySlotAction(key, val)


_Groups = dict[tuple[str, ty.Any], set[GroupingLeaf]]
_NonGroupLeaves = list[Leaf]


class GroupingSource(Source):
    def __init__(self, name: str, sources: list[Source]) -> None:
        Source.__init__(self, name)
        self.sources = sources

    def _get_groups(
        self, force_update: bool
    ) -> tuple[_Groups, _NonGroupLeaves]:
        groups: _Groups = defaultdict(set)
        non_group_leaves: list[Leaf] = []

        for src in self.sources:
            leaves = Source.get_leaves(src, force_update)
            for leaf in leaves or ():
                try:
                    slots = leaf.slots()  # type: ignore
                except AttributeError:  # noqa:PERF203
                    # Let through Non-grouping leaves
                    non_group_leaves.append(leaf)
                else:
                    assert isinstance(leaf, GroupingLeaf)

                    if not leaf.grouping_slots:
                        self.output_error(
                            "GroupingLeaf has no grouping slots", repr(leaf)
                        )
                        continue

                    for slot in leaf.grouping_slots:
                        if value := slots.get(slot):
                            groups[(slot, value)].add(leaf)

        return groups, non_group_leaves

    def _merge_groups(self, groups: _Groups) -> set[tuple[str, ty.Any]]:
        redundant_keys = set()

        def merge_groups(key1, key2):
            if groups[key1] is not groups[key2]:
                groups[key1].update(groups[key2])
                groups[key2] = groups[key1]
                redundant_keys.add(key2)

        # Find all (slot, value) combinations that have more than one leaf
        # and merge those groups
        for (slot, value), gleaves in groups.items():
            if len(gleaves) <= 1:
                continue

            for leaf in list(gleaves):
                for slot2 in leaf.grouping_slots:
                    for value2 in leaf.all(slot2):
                        if value2:
                            merge_groups((slot, value), (slot2, value2))

        return redundant_keys

    def get_leaves(self, force_update: bool = False) -> ty.Iterable[Leaf]:
        starttime = time.time()
        # map (slot, value) -> group
        groups, non_group_leaves = self._get_groups(force_update)

        # Keep track of keys that are only duplicate references
        redundant_keys = self._merge_groups(groups)

        keys = set(groups)
        keys.difference_update(redundant_keys)

        leaves: ty.Iterable[Leaf] = (
            self._make_group_leader(groups[K]) for K in keys
        )
        if self.should_sort_lexically():
            leaves = kupferstring.locale_sort(leaves)

        if (mergetime := time.time() - starttime) > 0.05:  # noqa: PLR2004
            self.output_debug(f"Warning(?): merged in {mergetime} seconds")

        return itertools.chain(non_group_leaves, leaves)

    def repr_key(self) -> ty.Any:
        # Distinguish when used as GroupingSource
        if isinstance(self, GroupingSource):
            return str(self)

        return Source.repr_key(self)

    @classmethod
    def _make_group_leader(cls, leaves: set[GroupingLeaf]) -> Leaf:
        if len(leaves) == 1:
            (leaf,) = leaves
            return leaf

        obj = copy.copy(next(iter(leaves)))
        obj.links = list(leaves)
        for other in leaves:
            obj.kupfer_add_alias(str(other))
            # adding the other's aliases can be misleading
            # since the matched email address might not be
            # what we are e-mailing
            # obj.name_aliases.update(other.name_aliases)
        return obj


class ToplevelGroupingSource(GroupingSource):
    """Sources of this type group their leaves with others in the toplevel
    of the catalog."""

    _sources: ty.ClassVar[
        dict[str, weakref.WeakKeyDictionary[Source, int]]
    ] = {}

    def __init__(self, name: str, category: str) -> None:
        GroupingSource.__init__(self, name, [self])
        self.category = category

    def toplevel_source(self) -> Source:
        if self.category not in self._sources:
            return self

        sources = list(self._sources[self.category].keys())
        return GroupingSource(self.category, sources)

    def initialize(self) -> None:
        if self.category not in self._sources:
            self._sources[self.category] = weakref.WeakKeyDictionary()

        self._sources[self.category][self] = 1
        self.output_debug(f"Register {self.category} source {self}")

    def finalize(self) -> None:
        del self._sources[self.category][self]
        self.output_debug(f"Unregister {self.category} source {self}")


class _GroupedItemsSource(Source):
    def __init__(self, leaf: GroupingLeaf) -> None:
        Source.__init__(self, str(leaf))
        self._leaf = leaf

    def get_items(self) -> ty.Iterator[Leaf]:
        yield from self._leaf.links

    def repr_key(self) -> ty.Any:
        return repr(self._leaf)


class CopySlotAction(Action):
    rank_adjust = -5

    """Action for GroupingLeaf - copy given slot value to clipboard."""

    def __init__(self, slot_name: str, value: str) -> None:
        name = slot_name.replace("_", " ").lower()
        super().__init__(_("Copy %s") % name)
        self.value = value

    def wants_context(self):
        return True

    def activate(
        self,
        leaf: Leaf,
        iobj: Leaf | None = None,
        ctx: commandexec.ExecutionToken | None = None,
    ) -> Leaf | None:
        assert ctx
        clip = Gtk.Clipboard.get_for_display(
            ctx.environment.get_screen().get_display(), Gdk.SELECTION_CLIPBOARD
        )
        clip.set_text(self.value, -1)
        return None

    def get_description(self) -> str:
        return _("Copy '%s' to clipboard") % self.value



__kupfer_name__ = _("Dictionary")
__kupfer_actions__ = ("LookUp",)
__description__ = _("Look up word in dictionary")
__version__ = "2023-04-14"
__author__ = "Ulrik, KB"

import shutil
import typing as ty
from dataclasses import dataclass

from kupfer import launch, plugin_support
from kupfer.desktop_launch import SpawnError
from kupfer.obj import Action, OperationError, TextLeaf

if ty.TYPE_CHECKING:
    from gettext import gettext as _


@dataclass
class Dict:
    title: str
    args: list[str]


dictionaries = {
    "gnome-dictionary": Dict(
        _("Gnome Dictionary"), ["gnome-dictionary", "--look-up=%s"]
    ),
    "mate-dictionary": Dict(
        _("Mate Dictionary"), ["mate-dictionary", "--look-up=%s"]
    ),
    "purple": Dict(_("Purple"), ["purple", "--define=%s"]),
    "xfce4-dict": Dict(_("Xfce4 Dict"), ["xfce4-dict", "--dict", "%s"]),
    "org.goldendict.GoldenDict": Dict(_("GoldenDict"), ["goldendict", "%s"]),
}

__kupfer_settings__ = plugin_support.PluginSettings(
    {
        "key": "dictionary",
        "label": _("Dictionary"),
        "type": str,
        "alternatives": [
            (key, dic.title) for key, dic in dictionaries.items()
        ],
        "value": "gnome-dictionary",
    }
)


class LookUp(Action):
    def __init__(self):
        Action.__init__(self, _("Look Up"))

    def activate(self, leaf, iobj=None, ctx=None):
        text = leaf.object
        dict_id = __kupfer_settings__["dictionary"]
        dict_def = dictionaries[dict_id]

        if not shutil.which(dict_def.args[0]):
            raise OperationError(f"{dict_def.title} not available")

        dict_argv = [arg.replace("%s", text) for arg in dict_def.args]
        try:
            launch.spawn_async_notify_as(dict_id + ".desktop", dict_argv)
        except SpawnError as exc:
            raise OperationError(exc) from exc

    def item_types(self):
        yield TextLeaf

    def valid_for_item(self, leaf):
        text = leaf.object
        return len(text.split("\n", 1)) <= 1

    def get_description(self):
        curr_dict = __kupfer_settings__["dictionary"]
        return _("Look up word in %s") % dictionaries[curr_dict].title

    def get_icon_name(self):
        return "accessories-dictionary"



"""DataController"""

from __future__ import annotations

import itertools
import os
import sys
import typing as ty
from collections import defaultdict
from contextlib import suppress
from enum import IntEnum
from pathlib import Path

from gi.repository import GLib, GObject

from kupfer.core import (
    commandexec,
    execfile,
    learn,
    pluginload,
    qfurl,
    search,
    settings,
)
from kupfer.core.panes import (
    LeafPane,
    Pane,
    PrimaryActionPane,
    SearchContext,
    SecondaryObjectPane,
)
from kupfer.core.sources import get_source_controller
from kupfer.obj import compose
from kupfer.obj import objects as kobjects
from kupfer.obj.base import (
    Action,
    ActionGenerator,
    AnySource,
    KupferObject,
    Leaf,
    Source,
    TextSource,
)
from kupfer.obj.filesrc import DirectorySource, FileSource
from kupfer.support import pretty, scheduler

if ty.TYPE_CHECKING:
    from kupfer.core.search import Rankable
    from kupfer.support.types import ExecInfo
    from kupfer.ui.uievents import GUIEnvironmentContext

__all__ = ("DataController", "PaneMode", "PaneSel")

DATA_SAVE_INTERVAL_S = 3660


# "Enums"
# Which pane
class PaneSel(IntEnum):
    SOURCE = 1
    ACTION = 2
    OBJECT = 3


# In two-pane or three-pane mode
class PaneMode(IntEnum):
    # two-panels mode: source -> action
    SOURCE_ACTION = 1
    # three panels mode: source -> action -> indirect object
    SOURCE_ACTION_OBJECT = 2
    # two-panels mode: action -> object (source); in this mode we cheat -
    # for specific leaf (ActionLeaf) which have only one action (ExecuteAction)
    # we by default select this action and hide panel with actions, so user
    # must select indirect object.
    ACTION_OBJECT = 3


# pylint: disable=too-many-public-methods
class DataController(GObject.GObject, pretty.OutputMixin):  # type:ignore
    """Sources <-> Actions controller.

    The data controller must be created before main program commences,
    so it can register itself at the scheduler correctly.
    """

    _instance: DataController | None = None

    @classmethod
    def instance(cls) -> DataController:
        """Get instance of DataController."""
        if cls._instance is None:
            cls._instance = DataController()

        return cls._instance

    __gtype_name__ = "DataController"

    def __init__(self):
        super().__init__()

        self._source_pane = LeafPane()
        self._source_pane.connect("new-source", self._on_new_source)
        self._object_pane = SecondaryObjectPane()
        self._object_pane.connect("new-source", self._on_new_source)
        self._action_pane = PrimaryActionPane()
        for pane, ctl in self._all_pane_ctl():
            ctl.connect("search-result", self._on_pane_search_result, pane)

        self._mode: PaneMode | None = None
        self._search_ids = itertools.count(1)
        self._latest_interaction = -1
        self._execution_context = (
            commandexec.default_action_execution_context()
        )
        self._execution_context.connect(
            "command-result", self._on_command_execution_result
        )
        self._execution_context.connect(
            "late-command-result", self._on_late_command_execution_result
        )

        self._save_data_timer = scheduler.Timer()

        sch = scheduler.get_scheduler()
        sch.connect("load", self._on_load)
        sch.connect("display", self._on_display)
        sch.connect("finish", self._on_finish)

    def _get_panectl(self, pane: PaneSel) -> Pane[Leaf] | Pane[Action]:
        if pane == PaneSel.SOURCE:
            return self._source_pane
        if pane == PaneSel.ACTION:
            return self._action_pane
        if pane == PaneSel.OBJECT:
            return self._object_pane

        raise ValueError(f"invalid pane {pane}")

    def _all_pane_ctl(
        self,
    ) -> ty.Iterator[tuple[PaneSel, Pane[Leaf] | Pane[Action]]]:
        yield PaneSel.SOURCE, self._source_pane
        yield PaneSel.ACTION, self._action_pane
        yield PaneSel.OBJECT, self._object_pane

    def _register_text_sources(
        self, plugin_id: str, srcs: ty.Iterable[TextSource]
    ) -> None:
        """Pass in text sources as @srcs

        we register text sources"""
        sctr = get_source_controller()
        sctr.add_text_sources(plugin_id, srcs)

    def _register_action_decorators(
        self, plugin_id: str, actions: list[Action]
    ) -> None:
        # Keep a mapping: Decorated Leaf Type -> List of actions
        decorate_types: defaultdict[type[Leaf], list[Action]] = defaultdict(
            list
        )
        for action in actions:
            for appl_type in action.item_types():
                decorate_types[appl_type].append(action)

        if not decorate_types:
            return

        sctr = get_source_controller()
        sctr.add_action_decorators(plugin_id, decorate_types)

    def _register_content_decorators(
        self, plugin_id: str, contents: ty.Collection[ty.Type[Source]]
    ) -> None:
        """Register the sequence of classes @contents as
        potential content decorators. Classes not conforming to
        the decoration protocol (most importantly, ``.decorates_type()``)
        will be skipped."""
        # Keep a mapping:
        # Decorated Leaf Type -> Set of content decorator types
        decorate_item_types: defaultdict[
            ty.Type[Leaf], set[ty.Type[Source]]
        ] = defaultdict(set)

        for content in contents:
            with suppress(AttributeError):
                applies = content.decorates_type()  # type: ignore
                decorate_item_types[applies].add(content)

        if not decorate_item_types:
            return

        sctr = get_source_controller()
        sctr.add_content_decorators(plugin_id, decorate_item_types)

    def _register_action_generators(
        self, plugin_id: str, generators: ty.Iterable[ActionGenerator]
    ) -> None:
        sctr = get_source_controller()
        for generator in generators:
            sctr.add_action_generator(plugin_id, generator)

    def _on_load(self, _sched: ty.Any) -> None:
        """Begin Data Controller work when we get application 'load' signal.

        Load the data model from saved configuration and caches
        """
        setctl = settings.get_settings_controller()
        setctl.connect("plugin-enabled-changed", self._on_plugin_enabled)
        setctl.connect(
            "plugin-toplevel-changed", self._on_plugin_catalog_changed
        )

        self._load_all_plugins()
        dir_src, indir_src = self._get_directory_sources()
        sctr = get_source_controller()
        sctr.add(None, dir_src, toplevel=True)
        sctr.add(None, indir_src, toplevel=False)
        sctr.initialize()
        learn.load()

    def _on_display(self, _sched: ty.Any) -> None:
        self._reload_source_root()
        self._save_data_timer.set(DATA_SAVE_INTERVAL_S, self._save_data)

    def _get_directory_sources(
        self,
    ) -> tuple[tuple[DirectorySource, ...], tuple[DirectorySource, ...]]:
        """Return a tuple of dir_sources, indir_sources for directory sources
        directly included and for catalog inclusion respectively."""
        setctl = settings.get_settings_controller()
        source_config = setctl.get_config
        dir_depth = source_config("DeepDirectories", "Depth")

        def file_source(opt, depth=1):
            abs_path = os.path.abspath(os.path.expanduser(opt))
            return FileSource([abs_path], depth)

        indir_sources = itertools.chain(
            (
                DirectorySource(item, toplevel=True)
                for item in setctl.get_directories(False)
                if Path(item).is_dir()
            ),
            (
                file_source(item, dir_depth)
                for item in source_config("DeepDirectories", "Catalog")
            ),
        )

        dir_sources = itertools.chain(
            (
                DirectorySource(item, toplevel=True)
                for item in setctl.get_directories(True)
                if Path(item).is_dir()
            ),
            (
                file_source(item, dir_depth)
                for item in source_config("DeepDirectories", "Direct")
            ),
        )

        return tuple(dir_sources), tuple(indir_sources)

    def _load_all_plugins(self):
        """Insert all plugin sources into the catalog."""
        # pylint: disable=import-outside-toplevel
        from kupfer.core import plugins

        setctl = settings.get_settings_controller()
        for item in sorted(plugins.get_plugin_ids()):
            if setctl.get_plugin_enabled(item):
                sources_ = self._load_plugin(item)
                self._insert_sources(item, sources_, initialize=False)

    def _load_plugin(self, plugin_id: str) -> set[Source]:
        """Load @plugin_id, register all its Actions, Content and TextSources.
        Return its sources."""
        with pluginload.exception_guard(plugin_id):
            plugin = pluginload.load_plugin(plugin_id)
            self._register_text_sources(plugin_id, plugin.text_sources)
            self._register_action_decorators(
                plugin_id, plugin.action_decorators
            )
            self._register_content_decorators(
                plugin_id, plugin.content_decorators
            )
            self._register_action_generators(
                plugin_id, plugin.action_generators
            )
            return set(plugin.sources)

        return set()

    def _on_plugin_enabled(
        self, _setctl: ty.Any, plugin_id: str, enabled: bool | int
    ) -> None:
        """Enable plugin, load if necessary."""
        # pylint: disable=import-outside-toplevel
        from kupfer.core import plugins

        if enabled and not plugins.is_plugin_loaded(plugin_id):
            srcs = self._load_plugin(plugin_id)
            self._insert_sources(plugin_id, srcs, initialize=True)

            # force update sources after plugin enable
            for src in srcs:
                src.mark_for_update()

        elif not enabled:
            self._remove_plugin(plugin_id)

    def _remove_plugin(self, plugin_id: str) -> None:
        sctl = get_source_controller()
        if sctl.remove_objects_for_plugin_id(plugin_id):
            self._reload_source_root()

        pluginload.remove_plugin(plugin_id)

    def _reload_source_root(self) -> None:
        """Reload items from root sources."""
        self.output_debug("Reloading source root")
        sctl = get_source_controller()
        if sctl.root:
            self._source_pane.source_rebase(sctl.root)

    def _on_plugin_catalog_changed(
        self, _setctl: ty.Any, _plugin_id: str, _toplevel: ty.Any
    ) -> None:
        self._reload_source_root()

    def _insert_sources(
        self,
        plugin_id: str,
        sources_: ty.Iterable[Source],
        initialize: bool = True,
    ) -> None:
        """Insert `sources_` into catalog. `Initialize` when True, initialize
        plugin and reload catalog root."""
        if not sources_:
            return

        sctl = get_source_controller()
        setctl = settings.get_settings_controller()
        for src in sources_:
            is_toplevel = setctl.get_source_is_toplevel(plugin_id, src)
            sctl.add(
                plugin_id, (src,), toplevel=is_toplevel, initialize=initialize
            )

        if initialize:
            self._reload_source_root()

    def _on_finish(self, _sched: ty.Any) -> None:
        "Close down the data model, save user data, and write caches to disk"
        get_source_controller().finalize()
        self._save_data(final_invocation=True)
        self.output_info("Saving cache...")
        get_source_controller().save_cache()

    def _save_data(self, final_invocation: bool = False) -> None:
        """Save Learning data and User's configuration data in sources
        (Recurring timer)."""
        self.output_info("Saving data...")
        learn.save()
        get_source_controller().save_data()
        if not final_invocation:
            self._save_data_timer.set(DATA_SAVE_INTERVAL_S, self._save_data)

    def _on_new_source(
        self, ctr: Pane[KupferObject], src: AnySource | None, select: ty.Any
    ) -> None:
        """Update leaf or secondary object panel on sources changes."""
        if not src:
            return

        if ctr is self._source_pane:
            pane = PaneSel.SOURCE
        elif ctr is self._object_pane:
            pane = PaneSel.OBJECT
        else:
            raise AttributeError

        root = ctr.is_at_source_root()
        self.emit("source-changed", pane, src, root, select)

    def reset(self) -> None:
        self._source_pane.reset()
        self._action_pane.reset()
        self._object_pane.reset()

    def soft_reset(self, pane: PaneSel) -> AnySource | None:
        if pane == PaneSel.ACTION:
            return None

        panectl: LeafPane = self._get_panectl(pane)
        return panectl.soft_reset()

    def cancel_search(self, pane: PaneSel | None = None) -> None:
        """Cancel any outstanding search, or the search for @pane"""
        panes = (
            (pane,)
            if pane
            else (PaneSel.SOURCE, PaneSel.ACTION, PaneSel.OBJECT)
        )
        for pane_ in panes:
            ctl = self._get_panectl(pane_)
            if ctl.outstanding_search > 0:
                GLib.source_remove(ctl.outstanding_search)
                ctl.outstanding_search = -1

    def search(
        self,
        pane: PaneSel,
        key: str = "",
        context: str | None = None,
        interactive: bool = False,
        lazy: bool = False,
        text_mode: bool = False,
    ) -> None:
        """Search: Register the search method in the event loop

        Will search in @pane's base using @key, promising to return
        @context in the notification about the result.

        if @interactive, the search result will return immediately
        if @lazy, will slow down search result reporting
        """
        self.cancel_search(pane)
        self._latest_interaction = self._execution_context.last_command_id
        ctl = self._get_panectl(pane)
        ctl.outstanding_search_id = next(self._search_ids)
        wrapcontext = (ctl.outstanding_search_id, context)
        if interactive:
            ctl.search(key, wrapcontext, text_mode)
            return

        timeout = 300 if lazy else 0 if not key else 50 // len(key)

        def ctl_search(*args):
            ctl.outstanding_search = -1
            return ctl.search(*args)

        ctl.outstanding_search = GLib.timeout_add(
            timeout, ctl_search, key, wrapcontext, text_mode
        )

    def _on_pane_search_result(
        self,
        panectl: Pane[ty.Any],
        match: Rankable | None,
        match_iter: ty.Iterable[Rankable],
        wrapcontext: SearchContext,
        pane: PaneSel,
    ) -> bool:
        search_id, context = wrapcontext
        if search_id == panectl.outstanding_search_id:
            self.emit("search-result", pane, match, match_iter, context)
            return False

        self.output_debug("Skipping late search", match, context)
        return True

    def select(self, pane: PaneSel, item: Leaf | Action | None) -> None:
        """Select @item in @pane to self-update relevant places"""
        # If already selected, do nothing
        panectl = self._get_panectl(pane)
        if item == panectl.get_selection():
            return

        self.cancel_search()
        if pane == PaneSel.SOURCE:
            assert isinstance(panectl, LeafPane)
            assert item is None or isinstance(item, Leaf)
            panectl.select(item)
            # populate actions
            citem = self._get_pane_object_composed(self._source_pane)
            self._action_pane.set_item(citem)
            self.search(PaneSel.ACTION, interactive=True)

            # change mode according to selected in first panel leaf
            newmode = self._mode
            if isinstance(item, kobjects.ActionLeaf):
                newmode = PaneMode.ACTION_OBJECT
            elif self._mode == PaneMode.ACTION_OBJECT:
                newmode = PaneMode.SOURCE_ACTION

            if newmode != self._mode:
                self._mode = newmode
                self.emit("mode-changed", self._mode, item)

            if self._mode == PaneMode.SOURCE_ACTION_OBJECT:
                self.object_stack_clear(PaneSel.OBJECT)
                self._populate_third_pane()

        elif pane == PaneSel.ACTION:
            assert isinstance(panectl, PrimaryActionPane)
            assert item is None or isinstance(item, Action), str(type(item))
            panectl.select(item)
            self.object_stack_clear(PaneSel.OBJECT)
            if item and item.requires_object():
                newmode = PaneMode.SOURCE_ACTION_OBJECT
            else:
                newmode = PaneMode.SOURCE_ACTION

            if newmode != self._mode:
                self._mode = newmode
                self.emit("mode-changed", self._mode, item)

            if self._mode == PaneMode.SOURCE_ACTION_OBJECT:
                self._populate_third_pane()
        else:  # OBJECT
            assert isinstance(panectl, SecondaryObjectPane)
            assert item is None or isinstance(item, Leaf)
            panectl.select(item)

    def _populate_third_pane(self) -> None:
        citem = self._get_pane_object_composed(self._source_pane)
        action = self._action_pane.get_selection()
        assert isinstance(action, Action)
        self._object_pane.set_item_and_action(citem, action)
        self.search(PaneSel.OBJECT, lazy=True)

    def get_can_enter_text_mode(self, pane: PaneSel) -> bool:
        panectl = self._get_panectl(pane)
        return panectl.get_can_enter_text_mode()

    def get_should_enter_text_mode(self, pane: PaneSel) -> bool:
        panectl = self._get_panectl(pane)
        return panectl.get_should_enter_text_mode()

    def validate(self) -> None:
        """Check if all selected items are still valid (for example after being
        spawned again, old item still focused).

        This will trigger .select() with None if items are not valid..
        """

        def valid_check(obj):
            return not (hasattr(obj, "is_valid") and not obj.is_valid())

        for pane, panectl in self._all_pane_ctl():
            sel = panectl.get_selection()
            if not valid_check(sel):
                self.emit("pane-reset", pane, None)
                self.select(pane, None)

            if self._has_object_stack(pane):
                new_stack = list(filter(valid_check, panectl.object_stack))
                if new_stack != panectl.object_stack:
                    self._set_object_stack(pane, new_stack)

    def browse_up(self, pane: PaneSel) -> bool:
        """Try to browse up to previous sources, from current source"""
        if pane == PaneSel.SOURCE:
            return self._source_pane.browse_up()

        if pane == PaneSel.OBJECT:
            return self._object_pane.browse_up()

        return False

    def browse_down(self, pane: PaneSel, alternate: bool = False) -> None:
        """Browse into @leaf if it's possible and save away the previous sources
        in the stack. If @alternate, use the Source's alternate method"""
        if pane == PaneSel.ACTION:
            return

        # record used object if we browse down
        panectl = self._get_panectl(pane)
        sel, key = panectl.get_selection(), panectl.get_latest_key()
        if panectl.browse_down(alternate=alternate):
            learn.record_search_hit(sel, key)

    def activate(self, ui_ctx: GUIEnvironmentContext) -> None:
        """Activate current selection.

        @ui_ctx: GUI environment context object
        """
        leaf, action, sobject = self._get_current_command_objects()

        # register search to learning database
        learn.record_search_hit(leaf, self._source_pane.get_latest_key())
        learn.record_search_hit(action, self._action_pane.get_latest_key())
        if sobject and self._mode == PaneMode.SOURCE_ACTION_OBJECT:
            learn.record_search_hit(
                sobject, self._object_pane.get_latest_key()
            )

        if not leaf or not action:
            return

        learn.record_action_activations(action, leaf)

        try:
            ctx = self._execution_context
            res, _ret = ctx.run(leaf, action, sobject, ui_ctx=ui_ctx)
        except commandexec.ActionExecutionError:
            self.output_exc()
            return

        if not res.is_sync:
            self.emit("launched-action", leaf, action, sobject)

    def execute_file(
        self,
        filepath: str,
        ui_ctx: GUIEnvironmentContext,
        on_error: ty.Callable[[ExecInfo], None],
    ) -> bool:
        """Execute a .kfcom file"""
        ctx = self._execution_context
        try:
            cmd_objs = execfile.parse_kfcom_file(filepath)
            assert len(cmd_objs) <= 3  # noqa:PLR2004
            ctx.run(*cmd_objs, ui_ctx=ui_ctx)  # type: ignore
            return True
        except commandexec.ActionExecutionError:
            self.output_exc()
        except execfile.ExecutionError:
            on_error(sys.exc_info())

        return False

    def _insert_object(self, pane: PaneSel, obj: Leaf) -> None:
        """Insert @obj in @pane: prepare the object, then emit pane-reset"""
        if pane == PaneSel.SOURCE:
            self._decorate_object(obj)

        self.emit("pane-reset", pane, search.wrap_rankable(obj))

    def _decorate_object(self, *objects: Leaf) -> None:
        sctl = get_source_controller()
        for obj in objects:
            sctl.decorate_object(obj)

    def insert_objects(self, pane: PaneSel, objects: list[Leaf]) -> None:
        """Select @objects in @pane"""
        if pane != PaneSel.SOURCE:
            raise ValueError("Can only insert in first pane")

        self._decorate_object(*objects[:-1])
        self._set_object_stack(pane, objects[:-1])
        self._insert_object(pane, objects[-1])

    def _on_command_execution_result(
        self,
        ctx: commandexec.ActionExecutionContext,
        result_type: commandexec.ExecResult | int,
        ret: ty.Any,
        uictx: GUIEnvironmentContext,
    ) -> None:
        result_type = commandexec.ExecResult(result_type)
        if result_type == commandexec.ExecResult.SOURCE:
            self.object_stack_clear_all()
            self._source_pane.push_source(ret)
        elif result_type == commandexec.ExecResult.OBJECT:
            self.object_stack_clear_all()
            self._insert_object(PaneSel.SOURCE, ret)
        elif result_type == commandexec.ExecResult.REFRESH:
            pass
        else:
            return

        self.emit("command-result", result_type, uictx)

    def _on_late_command_execution_result(
        self,
        ctx: commandexec.ActionExecutionContext,
        id_: int,
        result_type: commandexec.ExecResult | int,
        ret: ty.Any,
        uictx: GUIEnvironmentContext,
    ) -> None:
        """Receive late command result"""
        if self._latest_interaction < id_:
            self._on_command_execution_result(ctx, result_type, ret, uictx)

    def find_object(self, url: str) -> None:
        """Find object with URI @url and select it in the first pane"""
        sctrl = get_source_controller()
        qfu = qfurl.Qfurl(url=url)
        found = qfu.resolve_in_catalog(sctrl.get_sources())
        if found and found != self._source_pane.get_selection():
            self._insert_object(PaneSel.SOURCE, found)

    def mark_as_default(self, pane: PaneSel) -> None:
        """Make the object selected on @pane as default for the selection in
        previous pane."""
        if pane in (PaneSel.SOURCE, PaneSel.OBJECT):
            raise RuntimeError("Setting default on pane 1 or 3 not supported")

        obj = ty.cast("Leaf", self._source_pane.get_selection())
        act = ty.cast("Action", self._action_pane.get_selection())
        assert obj and act
        learn.set_correlation(act, obj)

    def get_object_has_affinity(self, pane: PaneSel) -> bool:
        """Return ``True`` if we have any recorded affinity for the object
        selected in @pane"""
        panectl = self._get_panectl(pane)
        if selection := panectl.get_selection():
            assert isinstance(selection, Leaf)
            return learn.get_object_has_affinity(selection)

        return False

    def erase_object_affinity(self, pane: PaneSel) -> None:
        """Erase all learned and configured affinity for the selection of
        @pane."""
        panectl = self._get_panectl(pane)
        if selection := panectl.get_selection():
            assert isinstance(selection, Leaf)
            learn.erase_object_affinity(selection)

    def compose_selection(self) -> None:
        leaf, action, iobj = self._get_current_command_objects()
        if leaf is None:
            return

        self.object_stack_clear_all()
        assert action
        obj = compose.ComposedLeaf(leaf, action, iobj)
        self._insert_object(PaneSel.SOURCE, obj)

    def _get_pane_object_composed(self, pane: Pane[Leaf]) -> Leaf | None:
        objects: list[Leaf] = list(pane.object_stack)
        sel = pane.get_selection()
        if sel and sel not in objects:
            objects.append(sel)

        if not objects:
            return None

        if len(objects) == 1:
            return objects[0]

        return compose.MultipleLeaf(objects)

    def _get_current_command_objects(
        self,
    ) -> tuple[Leaf, Action, Leaf | None] | tuple[None, None, None]:
        """Return a tuple of current (obj, action, iobj)."""
        objects = self._get_pane_object_composed(self._source_pane)
        action: Action | None = self._action_pane.get_selection()
        if objects is None or action is None:
            return (None, None, None)

        iobjects = self._get_pane_object_composed(self._object_pane)
        if self._mode in (
            PaneMode.SOURCE_ACTION_OBJECT,
            PaneMode.ACTION_OBJECT,
        ):
            if not iobjects:
                return (None, None, None)

        else:
            iobjects = None

        return (objects, action, iobjects)

    def _has_object_stack(self, pane: PaneSel) -> bool:
        return pane in (PaneSel.SOURCE, PaneSel.OBJECT)

    def _set_object_stack(self, pane: PaneSel, newstack: list[Leaf]) -> None:
        panectl = self._get_panectl(pane)
        panectl.object_stack = newstack
        self.emit("object-stack-changed", pane)

    def object_stack_push(self, pane: PaneSel, object_: KupferObject) -> bool:
        """Push @object_ onto the stack."""
        if not self._has_object_stack(pane):
            return False

        panectl = self._get_panectl(pane)
        if object_ not in panectl.object_stack:
            panectl.object_stack_push(object_)
            self.emit("object-stack-changed", pane)

        return True

    def object_stack_pop(self, pane: PaneSel) -> bool:
        if not self._has_object_stack(pane):
            return False

        panectl = self._get_panectl(pane)
        obj = panectl.object_stack_pop()
        self._insert_object(pane, obj)
        self.emit("object-stack-changed", pane)
        return True

    def object_stack_clear(self, pane: PaneSel) -> None:
        if not self._has_object_stack(pane):
            return

        panectl = self._get_panectl(pane)
        panectl.object_stack.clear()
        self.emit("object-stack-changed", pane)

    def object_stack_clear_all(self) -> None:
        """Clear the object stack for all panes."""
        # action don't have stack
        # self.object_stack_clear(PaneSel.ACTION)
        self.object_stack_clear(PaneSel.OBJECT)
        self.object_stack_clear(PaneSel.SOURCE)

    def get_object_stack(self, pane: PaneSel) -> list[Leaf]:
        if not self._has_object_stack(pane):
            return []

        panectl = self._get_panectl(pane)
        assert isinstance(panectl, (LeafPane, SecondaryObjectPane))
        return panectl.object_stack


# pane cleared or set with new item
# pane, item
GObject.signal_new(
    "pane-reset",
    DataController,
    GObject.SignalFlags.RUN_LAST,
    GObject.TYPE_BOOLEAN,
    (
        GObject.TYPE_INT,
        GObject.TYPE_PYOBJECT,
    ),
)

# pane, match, iter to matches, context
GObject.signal_new(
    "search-result",
    DataController,
    GObject.SignalFlags.RUN_LAST,
    GObject.TYPE_BOOLEAN,
    (
        GObject.TYPE_INT,
        GObject.TYPE_PYOBJECT,
        GObject.TYPE_PYOBJECT,
        GObject.TYPE_PYOBJECT,
    ),
)

GObject.signal_new(
    "source-changed",
    DataController,
    GObject.SignalFlags.RUN_LAST,
    GObject.TYPE_BOOLEAN,
    (int, object, bool, GObject.TYPE_PYOBJECT),
)

# mode, None(?)
GObject.signal_new(
    "mode-changed",
    DataController,
    GObject.SignalFlags.RUN_LAST,
    GObject.TYPE_BOOLEAN,
    (
        GObject.TYPE_INT,
        GObject.TYPE_PYOBJECT,
    ),
)

# object stack update signal
# arguments: pane
GObject.signal_new(
    "object-stack-changed",
    DataController,
    GObject.SignalFlags.RUN_LAST,
    GObject.TYPE_BOOLEAN,
    (GObject.TYPE_INT,),
)
# when an command returned a result
# arguments: result type, gui_context
GObject.signal_new(
    "command-result",
    DataController,
    GObject.SignalFlags.RUN_LAST,
    GObject.TYPE_BOOLEAN,
    (GObject.TYPE_INT, GObject.TYPE_PYOBJECT),
)

# when an action was launched
# arguments: leaf, action, secondary leaf
GObject.signal_new(
    "launched-action",
    DataController,
    GObject.SignalFlags.RUN_LAST,
    GObject.TYPE_BOOLEAN,
    (GObject.TYPE_PYOBJECT, GObject.TYPE_PYOBJECT, GObject.TYPE_PYOBJECT),
)



__kupfer_name__ = _("Higher-order Actions")
__kupfer_actions__ = ("Select", "TakeResult", "DiscardResult")
__description__ = _("Tools to work with commands as objects")
__version__ = "2010-01-11"
__author__ = "Ulrik Sverdrup <ulrik.sverdrup@gmail.com>"

import typing as ty

from kupfer.core import commandexec
from kupfer.obj import Action, Leaf, Source
from kupfer.obj.compose import ComposedLeaf, MultipleLeaf
from kupfer.support import pretty

if ty.TYPE_CHECKING:
    from gettext import gettext as _


class Select(Action):
    rank_adjust = -15

    def __init__(self):
        Action.__init__(self, _("Select in Kupfer"))

    def has_result(self):
        return True

    def activate(self, leaf, iobj=None, ctx=None):
        return leaf

    def item_types(self):
        yield Leaf


def _exec_no_show_result(composedleaf):
    pretty.print_debug(__name__, "Evaluating command", composedleaf)
    _obj, action, _iobj = composedleaf.object
    ret: commandexec.ActionResult = commandexec.activate_action(  # type: ignore
        None, *composedleaf.object
    )
    result_type = commandexec.parse_action_result(action, ret)
    if result_type == commandexec.ExecResult.OBJECT:
        return ret

    if result_type == commandexec.ExecResult.SOURCE:
        assert isinstance(ret, Source)
        leaves = list(ret.get_leaves())
        if not leaves:
            return None

        if len(leaves) == 1:
            return leaves[0]

        return MultipleLeaf(leaves)

    return None


def _save_result(cleaf):
    # Save the result of @cleaf into a ResultObject
    # When the ResultObject is to be restored from serialized state,
    # @cleaf is executed again.
    # NOTE: This will have unintended consequences outside Trigger use.
    leaf = _exec_no_show_result(cleaf)
    if leaf is None:
        return None

    class ResultObject(Leaf):
        serializable = 1

        def __init__(self, leaf, cleaf):
            Leaf.__init__(self, leaf.object, str(leaf))
            vars(self).update(vars(leaf))
            self.name = _("Result of %s (%s)") % (cleaf, self)
            self.__composed_leaf = cleaf
            self.__class__.__bases__ = (leaf.__class__, Leaf)

        def get_gicon(self):
            return None

        def get_icon_name(self):
            return Leaf.get_icon_name(self)

        def __reduce__(self):
            return (_save_result, (self.__composed_leaf,))

    return ResultObject(leaf, cleaf)


class TakeResult(Action):
    def __init__(self):
        Action.__init__(self, _("Run (Take Result)"))

    def has_result(self):
        return True

    def activate(self, leaf, iobj=None, ctx=None):
        return _save_result(leaf)

    def item_types(self):
        yield ComposedLeaf

    def valid_for_item(self, leaf):
        action = leaf.object[1]
        return (
            action.has_result() or action.is_factory()
        ) and not action.wants_context()

    def get_description(self):
        return _("Take the command result as a proxy object")


class DiscardResult(Action):
    """Run ComposedLeaf without showing the result"""

    def __init__(self):
        Action.__init__(self, _("Run (Discard Result)"))

    def wants_context(self):
        return True

    def activate(self, leaf, iobj=None, ctx=None):
        assert ctx
        commandexec.activate_action(ctx, *leaf.object)

    def item_types(self):
        yield ComposedLeaf

    def valid_for_item(self, leaf):
        action = leaf.object[1]
        return action.has_result() or action.is_factory()

    def get_description(self):
        return None



"""
Sources definition.

This file is a part of the program kupfer, which is
released under GNU General Public License v3 (or any later version),
see the main program file, and COPYING for details.
"""

from __future__ import annotations

import typing as ty

from kupfer.obj.base import Leaf, Source
from kupfer.obj.objects import SourceLeaf
from kupfer.support import itertools

if ty.TYPE_CHECKING:
    from gettext import gettext as _

__all__ = (
    "MultiSource",
    "SourcesSource",
)


class SourcesSource(Source):
    """A source whose items are SourceLeaves for @source"""

    def __init__(
        self,
        sources: ty.Collection[Source],
        name: str | None = None,
        use_reprs: bool = True,
    ) -> None:
        super().__init__(name or _("Catalog Index"))
        self.sources = sources
        self.use_reprs = use_reprs

    def get_items(self) -> ty.Iterable[Leaf]:
        """Ask each Source for a Leaf substitute, else yield a SourceLeaf"""
        if self.use_reprs:
            for src in self.sources:
                yield src.get_leaf_repr() or SourceLeaf(src)

            return

        yield from map(SourceLeaf, self.sources)

    def should_sort_lexically(self) -> bool:
        return True

    def get_description(self) -> str:
        return _("An index of all available sources")

    def get_icon_name(self) -> str:
        return "kupfer-catalog"


class MultiSource(Source):
    """A source whose items are the combined items of all @sources"""

    fallback_icon_name = "kupfer-catalog"

    def __init__(self, sources: ty.Collection[Source]) -> None:
        super().__init__(_("Catalog"))
        self.sources = sources

    def is_dynamic(self) -> bool:
        """MultiSource should be dynamic so some of its content also can be."""
        return True

    def get_items(self) -> ty.Iterable[Leaf]:
        uniq_srcs = itertools.unique_iterator(
            S.toplevel_source() for S in self.sources
        )
        for src in uniq_srcs:
            yield from src.get_leaves() or ()

    def get_description(self) -> str:
        return _("Root catalog")



"""
This is a simple plugin demonstration, how to add single, simple actions
"""

__kupfer_name__ = _("Wikipedia")
__kupfer_sources__ = ()
__kupfer_actions__ = ("WikipediaSearch",)
__description__ = _("Search in Wikipedia")
__version__ = "2017.1"
__author__ = "US"

import typing as ty
import urllib.parse

from kupfer import launch, plugin_support
from kupfer.obj import Action, Leaf, Source, TextLeaf

if ty.TYPE_CHECKING:
    from gettext import gettext as _

__kupfer_settings__ = plugin_support.PluginSettings(
    {
        "key": "lang",
        "label": _("Wikipedia languages (;-separated)"),
        "type": str,
        # TRANS: Default wikipedia language code
        "value": _("en"),
    },
)


class Lang(Leaf):
    def __init__(self, lang):
        super().__init__(lang, name=f"{lang}.wikipedia.org")

    def get_icon_name(self):
        return self.object


class LangSource(Source):
    def __init__(self, languages):
        super().__init__("Languages")
        self.languages = languages

    def get_items(self):
        return map(Lang, self.languages)


class WikipediaSearch(Action):
    action_accelerator = "w"

    def __init__(self):
        Action.__init__(self, _("Search in Wikipedia"))
        self._update_settings()

    def initialize(self):
        __kupfer_settings__.connect_settings_changed_cb(self._update_settings)
        self._update_settings()

    def _update_settings(self, *_args):
        languages: tuple[str, ...] = tuple(
            filter(
                None,
                (
                    lang.strip()
                    for lang in __kupfer_settings__["lang"].split(";")
                ),
            )
        )

        if not languages:
            languages = (_("en"),)

        self._languages = languages

    def activate(self, leaf, iobj=None, ctx=None):
        # Send in UTF-8 encoding
        lang_code = iobj.object if iobj else self._languages[0]
        search_url = f"https://{lang_code}.wikipedia.org/w/index.php?title=Special:Search&go=Go&"
        # will encode search=text, where `text` is escaped
        query_url = search_url + urllib.parse.urlencode(
            {"search": leaf.object}
        )
        launch.show_url(query_url)

    def item_types(self):
        yield TextLeaf

    def get_description(self):
        if len(self._languages) == 1:
            lang_code = self._languages[0]
            return _("Search for this term in %s.wikipedia.org") % lang_code

        return _("Search for this term in %s.wikipedia.org") % ""

    def get_icon_name(self):
        return "edit-find"

    def requires_object(self):
        return len(self._languages) > 1

    def object_source(self, for_item=None):
        return LangSource(self._languages)

    def object_types(self):
        yield Lang



#! /usr/bin/env python3
"""
Plugin generate recent opened files for vim/gvim.
"""

from __future__ import annotations

__kupfer_name__ = _("Vim")
__kupfer_sources__ = ("VimRecentsSource", "VimWikiSource")
__description__ = _("Load recent files edited in VIM/GVIM")
__version__ = "2023-04-02"
__author__ = "Karol Będkowski <karol.bedkowski@gmail.com>"

import os
import shutil
import typing as ty
from pathlib import Path

from kupfer import icons, launch, plugin_support
from kupfer.obj import Action, FileLeaf, Source, SourceLeaf
from kupfer.obj.apps import AppLeafContentMixin
from kupfer.obj.helplib import FileMonitorToken, FilesystemWatchMixin
from kupfer.support.datatools import simple_cache

if ty.TYPE_CHECKING:
    from gettext import gettext as _

    from gi.repository import Gio


plugin_support.check_any_command_available("vim", "gvim")

__kupfer_settings__ = plugin_support.PluginSettings(
    {
        "key": "wikis",
        "label": _("VimWiki directories:"),
        "type": list,
        "value": [],
        "helper": "choose_directory",
    },
    {
        "key": "limit",
        "label": _("Max recent documents:"),
        "type": int,
        "value": 25,
        "min": 1,
        "max": 100,
    },
)


@simple_cache
def _load_recent_files(viminfo: Path, limit: int, stamp: float) -> list[Path]:
    # stamp is file modification timestamp used to evict old cache items
    files = []
    with viminfo.open("rt", encoding="UTF-8", errors="replace") as fin:
        for line in fin:
            if not line.startswith("> "):
                continue

            if filepath := line[2:].strip():
                files.append(Path(filepath).expanduser())
                limit -= 1
                if not limit:
                    break

    return files


class VimRecentsSource(AppLeafContentMixin, Source, FilesystemWatchMixin):
    appleaf_content_id = ("vim", "gvim")
    source_scan_interval: int = 3600

    def __init__(self, name=None):
        super().__init__(name=_("Vim Recent Documents"))
        self._monitor_token: FileMonitorToken | None = None
        self._viminfo = Path("~/.viminfo").expanduser()

    def initialize(self):
        self._monitor_token = self.monitor_files(self._viminfo)

    def finalize(self):
        self.stop_monitor_fs_changes(self._monitor_token)

    def monitor_include_file(self, gfile: Gio.File) -> bool:
        return bool(gfile) and gfile.get_basename() == ".viminfo"

    def get_items_forced(self):
        try:
            _load_recent_files.cache_clear()
        except AttributeError:
            _load_recent_files.__wrapped__.cache_clear()  # type:ignore

        return self.get_items()

    def get_items(self):
        viminfo = self._viminfo
        if not viminfo.exists():
            self.output_debug("Viminfo not found at", viminfo)
            return ()

        limit = __kupfer_settings__["limit"]
        try:
            return map(
                FileLeaf,
                _load_recent_files(viminfo, limit, viminfo.stat().st_mtime),
            )
        except OSError:
            self.output_exc()

        return ()

    def get_icon_name(self):
        return "document-open-recent"

    def provides(self):
        yield FileLeaf


class VimWikiFile(FileLeaf):
    def __init__(self, path: Path, name: str, wikiname: str) -> None:
        super().__init__(path, name)
        self.kupfer_add_alias(f"vimwiki {wikiname} {name}")

    def get_actions(self):
        yield OpenVimWiki()


class VimWiki(SourceLeaf):
    def __init__(self, source: VimWikiFilesSource, name: str) -> None:
        super().__init__(source, name)

    def get_gicon(self):
        return icons.ComposedIconSmall("gvim", "emblem-documents")

    def get_description(self):
        return _("VimWiki in %s") % self.object.wikipath


class VimWikiSource(Source):
    def __init__(self):
        super().__init__(name=_("VimWiki Wikis"))

    def get_items(self):
        existing_wiki = []
        for path in __kupfer_settings__["wikis"] or ():
            filepath = Path(path).expanduser().resolve()
            if filepath.is_dir():
                name = filepath.name
                # make names unique (simple and not perfect)
                if name in existing_wiki:
                    name = f"{name} ({filepath.parent.name})"

                existing_wiki.append(name)
                yield VimWiki(VimWikiFilesSource(filepath), name)

    def provides(self):
        yield VimWiki

    def get_gicon(self):
        return icons.ComposedIconSmall("gvim", "emblem-documents")


class VimWikiFilesSource(Source, FilesystemWatchMixin):
    source_scan_interval: int = 3600

    def __init__(self, wikipath: Path) -> None:
        super().__init__(wikipath.name)
        self.wikipath = wikipath
        self._viminfo = Path("~/.viminfo").expanduser()
        self._monitor_token: FileMonitorToken | None = None

    def initialize(self):
        self._monitor_token = self.monitor_files(self._viminfo)

    def finalize(self):
        self.stop_monitor_fs_changes(self._monitor_token)

    def monitor_include_file(self, gfile: Gio.File) -> bool:
        if not gfile or gfile.get_basename() != ".viminfo":
            return False

        viminfo = self._viminfo
        if not viminfo.exists():
            return False

        # check is last modified files in vim belong do this wiki
        limit = __kupfer_settings__["limit"]
        for fname in _load_recent_files(
            viminfo, limit, viminfo.stat().st_mtime
        ):
            if fname.is_relative_to(self.wikipath):
                return True

        return False

    def mark_for_update(self, postpone=False):
        super().mark_for_update(postpone=True)

    def repr_key(self) -> str:
        return str(self.wikipath)

    def get_items(self) -> ty.Iterable[VimWikiFile]:
        # wiki can keep various files (txt, md, etc). Instead of read index
        # and guess extension - load add files in wiki
        wikipath = self.wikipath
        wikiname = self.wikipath.name
        for dirname, dirs, files in os.walk(wikipath):
            dirp = Path(dirname)
            # skip hidden directories
            if dirp.name[0] == ".":
                dirs.clear()
                continue

            for file in files:
                # skip hidden files
                if file[0] == ".":
                    continue

                filepath = dirp.joinpath(file)
                yield VimWikiFile(
                    filepath, str(filepath.relative_to(wikipath)), wikiname
                )

    def provides(self):
        yield VimWikiFile

    def should_sort_lexically(self):
        return True


class OpenVimWiki(Action):
    rank_adjust = 20

    def __init__(self):
        super().__init__(name=_("Open in Vim"))

    def activate(self, leaf, iobj=None, ctx=None):
        if cmd := shutil.which("gvim"):
            launch.spawn_async([cmd, leaf.object])
        elif cmd := shutil.which("vim"):
            launch.spawn_in_terminal([cmd, leaf.object])

    def get_icon_name(self) -> str:
        return "gvim"



from __future__ import annotations

__kupfer_name__ = _("Chromium Bookmarks")
__kupfer_sources__ = ("BookmarksSource",)
__description__ = _("Index of Chromium bookmarks")
__version__ = ""
__author__ = "Francesco Marella <francesco.marella@gmail.com>"

import typing as ty

from kupfer import config, plugin_support
from kupfer.obj import Source, UrlLeaf
from kupfer.obj.apps import AppLeafContentMixin
from kupfer.obj.helplib import FileMonitorToken, FilesystemWatchMixin
from kupfer.plugin import chromium_support

if ty.TYPE_CHECKING:
    from gettext import gettext as _


plugin_support.check_any_command_available("chromium", "chromium-browser")


def _get_chrome_conf_filepath() -> str | None:
    fpath = config.get_config_file("Bookmarks", package="chromium/Default")

    # If there is no chromium bookmarks file, look for a google-chrome one
    if not fpath:
        fpath = config.get_config_file(
            "Bookmarks", package="google-chrome/Default"
        )

    return fpath


class BookmarksSource(AppLeafContentMixin, Source, FilesystemWatchMixin):
    appleaf_content_id = ("chromium-browser", "chromium")
    source_scan_interval: int = 3600

    def __init__(self) -> None:
        self.monitor_token: FileMonitorToken | None = None
        super().__init__(_("Chromium Bookmarks"))

    def initialize(self) -> None:
        if fpath := _get_chrome_conf_filepath():
            self.monitor_token = self.monitor_directories(fpath)

    def monitor_include_file(self, gfile):
        return gfile and gfile.get_basename() == "lock"

    def mark_for_update(self, postpone=False):
        super().mark_for_update(postpone=True)

    def _get_chromium_items(self, fpath: str) -> ty.Iterator[UrlLeaf]:
        """Parse Chromium' bookmarks backups"""
        self.output_debug("Parsing", fpath)
        bookmarks = chromium_support.get_bookmarks(fpath)
        for book in bookmarks:
            yield UrlLeaf(book["url"], book["name"])

    def get_items(self) -> ty.Iterable[UrlLeaf]:
        if fpath := _get_chrome_conf_filepath():
            try:
                return self._get_chromium_items(fpath)
            except Exception as exc:
                self.output_error(exc)

        self.output_error("No Chromium bookmarks file found")
        return ()

    def get_description(self):
        return _("Index of Chromium bookmarks")

    def get_icon_name(self):
        return "chromium-browser"

    def provides(self):
        yield UrlLeaf



from __future__ import annotations

import itertools
from contextlib import suppress

__all__ = ("extract_title_body", "parse_time_interval")


def _unicode_truncate(ustr: str, length: int, encoding: str = "UTF-8") -> str:
    """Truncate @ustr to specific encoded byte length"""
    bstr = ustr.encode(encoding)[:length]
    return bstr.decode(encoding, "ignore")


def _split_first_line(text: str) -> tuple[str, str]:
    """Take first non-empty line of @text and rest"""
    lines, *rest = text.lstrip().split("\n", maxsplit=1)
    return lines, rest[0] if rest else ""


def _split_first_words(text: str, maxlen: int) -> tuple[str, str]:
    text = text.lstrip()
    first_text = _unicode_truncate(text, maxlen)
    words = first_text.split()
    if len(words) > 3:  # noqa: PLR2004
        first_words = " ".join(words[:-2])
        if text.startswith(first_words):
            first_text = first_words

    rest_text = text[len(first_text) :]
    return first_text, rest_text


def extract_title_body(text: str, maxtitlelen: int = 60) -> tuple[str, str]:
    """Prepare @text: Return a (title, body) tuple

    @text: A user-submitted paragraph or otherwise snippet of text. We
    try to detect an obvious title and then return the title and the
    following body. Otherwise we extract a title from the first words,
    and return the full text as body.

    @maxtitlelen: A unitless measure of approximate length of title.
    The default value yields a resulting title of approximately 60 ascii
    characters, or 20 asian characters.

    >>> extract_title_body("Short Text")
    ('Short Text', '')

    >>> title, body = extract_title_body(u"執筆方針については、項目名の付け方、"
    ...     "フォーマットや表記上の諸問題に関して多くの方針が存在している。")
    >>> print(title)
    執筆方針については、項目名の付け方、フォ
    >>> print(body)         # doctest: +ELLIPSIS
    執筆方針については、項目名の付け方、フォ...して多くの方針が存在している。
    """
    # if you don't make real tests, it's not not worth doing it at all.

    if not text.strip():
        return text, ""

    firstline, rest = _split_first_line(text)
    if len(firstline.encode("UTF-8")) <= maxtitlelen:
        return firstline, rest

    # We use the UTF-8 encoding and truncate due to it:
    # this is a good heuristic for ascii vs "wide characters"
    # it results in taking fewer characters if they are asian, which
    # is exactly what we want
    firstline, rest = _split_first_words(text, maxtitlelen)

    if rest.strip():
        return firstline, text

    return text, ""


def parse_time_interval(tstr: str) -> int:
    """Parse a time interval in `tstr`, return whole number of seconds.

    >>> parse_time_interval("2")
    2
    >>> parse_time_interval("1h 2m 5s")
    3725
    >>> parse_time_interval("2 min")
    120
    """
    weights = {
        "s": 1,
        "sec": 1,
        "m": 60,
        "min": 60,
        "h": 3600,
        "hours": 3600,
    }

    with suppress(ValueError):
        return int(tstr)

    total = 0
    amount = 0
    # Split the string in runs of digits and runs of characters
    for isdigit, group in itertools.groupby(tstr, str.isdigit):
        if not (part := "".join(group).strip()):
            continue

        if isdigit:
            amount = int(part)
        else:
            total += amount * weights.get(part.lower(), 0)
            amount = 0

    return total


if __name__ == "__main__":
    import doctest

    doctest.testmod()



from __future__ import annotations

import operator
import os
import typing as ty
from contextlib import suppress
from pathlib import Path

import gi
from gi.repository import Gdk, Gio, GLib, GObject, Gtk, Pango
from xdg import BaseDirectory as xdgBase
from xdg import DesktopEntry
from xdg import Exceptions as xdgExc

from kupfer import config, icons, launch, plugin_support, version
from kupfer.core import plugins, relevance, settings, sources
from kupfer.support import kupferstring, pretty, scheduler
from kupfer.ui import _widgets as widgets
from kupfer.ui import accelerators, getkey_dialog, keybindings, kupferhelp
from kupfer.ui.credentials_dialog import ask_user_credentials

if ty.TYPE_CHECKING:
    from gettext import gettext as _

    from kupfer.ui.uievents import GUIEnvironmentContext


# index in GtkNotebook
_PLUGIN_LIST_PAGE: ty.Final = 2
# List icon pixel size
_LIST_ICON_SIZE: ty.Final = 18

_KEYBINDING_TARGETS: ty.Final[dict[str, int]] = {
    "keybinding": keybindings.KEYBINDING_TARGET_DEFAULT,
    "magickeybinding": keybindings.KEYBINDING_TARGET_MAGIC,
}

_KUPFER_DESKTOP: ty.Final = "kupfer.desktop"
_AUTOSTART_KEY: ty.Final = "X-GNOME-Autostart-enabled"
_HIDDEN_KEY: ty.Final = "Hidden"


def _set_combobox(value: ty.Any, combobox: Gtk.ComboBoxText) -> None:
    """Set activate the alternative in the combobox with text value."""
    value = str(value)
    col = combobox.get_entry_text_column()
    for row in combobox.get_model():
        if row[col] == value:
            combobox.set_active_iter(row.iter)
            return


def _set_combobox_id(id_: ty.Any, combobox: Gtk.ComboBox) -> None:
    """Set activate the alternative in the combobox with id."""
    if id_:
        combobox.set_active_id(str(id_))


def _make_combobox_model(combobox: Gtk.ComboBox) -> None:
    # List store with columns (Name, ID)
    combobox_store = Gtk.ListStore(GObject.TYPE_STRING, GObject.TYPE_STRING)
    combobox.set_model(combobox_store)
    combobox_cell = Gtk.CellRendererText()
    combobox.pack_start(combobox_cell, True)
    combobox.add_attribute(combobox_cell, "text", 0)


def _create_plugin_change_cb(
    plugin_id: str,
    key: str,
    value_type: ty.Type[ty.Any],
    get_attr: str,
    no_false_values: bool = False,
) -> ty.Callable[[Gtk.Widget], None]:
    """Callback factory for the plugin parameter configuration.

    Callback load value from widget calling `get_attr` method and update
    setting in SettingsController.
    When `no_false_values`, ignore empty/0 values.
    """

    def callback(widget: Gtk.Widget) -> None:
        value = getattr(widget, get_attr)()
        if no_false_values and not value:
            return

        setctl = settings.get_settings_controller()
        setctl.set_plugin_config(plugin_id, key, value, value_type)

    return callback


def _create_plugin_credentials_cb(
    plugin_id: str, key: str
) -> ty.Callable[[Gtk.Widget], None]:
    """Create callback for credentials widget."""

    # TODO: check, not used / not working probably
    def callback(widget):
        setctl = settings.get_settings_controller()
        val_type = plugin_support.UserNamePassword
        # pylint: disable=no-member
        backend_name = (
            plugin_support.UserNamePassword.get_backend_name()  # type:ignore
        )
        assert backend_name
        # pylint: disable=no-member
        if (
            plugin_support.UserNamePassword.is_backend_encrypted()  # type:ignore
        ):
            information = (
                _("Using encrypted password storage: %s") % backend_name
            )
        else:
            information = _("Using password storage: %s") % backend_name

        upass = (
            setctl.get_plugin_config(plugin_id, key, val_type)
            or plugin_support.UserNamePassword()
        )

        # pylint: disable=no-member
        user_password = ask_user_credentials(
            upass.username,
            upass.password,
            information,
        )
        if user_password:
            # pylint: disable=no-member
            upass.username, upass.password = user_password  # type:ignore
            # TODO: fix
            setctl.set_plugin_config(
                plugin_id,
                key,
                upass,  # type:ignore
                val_type,
            )

    return callback


def _make_plugin_sett_widget_str(
    plugin_id: str,
    setting: str,
    plugin_settings: plugin_support.PluginSettings,
) -> tuple[Gtk.Widget, bool]:
    """Create widgets for str-like plugin setting.

    Usually this setting is simple Entry, but this may be changed by parameters:

    - alternatives: create combobox
    - multiline: create TextView
    - helper (choose_directory, choose_file): text + buttons for select
         directory/file.

    Returns: (widget, <is bigger widget>).
        Information about size of widget is used to better place of label.
    """
    wid = None
    multiline = False
    if alternatives := plugin_settings.get_alternatives(setting):
        wid = _make_plugin_sett_widget_combo(
            plugin_id, setting, plugin_settings, alternatives
        )

    elif plugin_settings.get_parameter(setting, "multiline"):
        wid = _make_plugin_sett_widget_multiline(
            plugin_id, setting, plugin_settings
        )
        multiline = True

    elif helper := plugin_settings.get_parameter(setting, "helper"):
        if helper in ("choose_directory", "choose_file"):
            wid = widgets.FileDirSelectWidget(
                plugin_id, setting, plugin_settings, helper
            )
            multiline = True
        else:
            pretty.print_error("unknown helper", helper)

    if not wid:
        wid = Gtk.Entry()
        wid.set_text(plugin_settings[setting])
        wid.set_hexpand(True)
        wid.set_size_request(100, 10)
        wid.connect(
            "changed",
            _create_plugin_change_cb(
                plugin_id,
                setting,
                str,
                "get_text",
                no_false_values=True,
            ),
        )

    return wid, multiline


def _make_plugin_sett_widget_combo(
    plugin_id: str,
    setting: str,
    plugin_settings: plugin_support.PluginSettings,
    alternatives: ty.Iterable[str],
) -> Gtk.Widget:
    """Create combobox widget for str-like plugin settings with alternatives."""
    wid = Gtk.ComboBoxText.new()
    wid.set_vexpand(False)
    val = plugin_settings[setting]
    active_index = -1
    for idx, text in enumerate(alternatives):
        id_ = text
        if isinstance(text, (tuple, list)):
            id_, text = text  # noqa: PLW2901

        wid.append(id=id_, text=text)
        if id_ == val:
            active_index = idx

    if active_index < 0:
        wid.prepend_text(val)
        active_index = 0

    wid.set_active(active_index)
    wid.connect(
        "changed",
        _create_plugin_change_cb(plugin_id, setting, str, "get_active_id"),
    )
    return wid


def _make_plugin_sett_widget_multiline(
    plugin_id: str,
    setting: str,
    plugin_settings: plugin_support.PluginSettings,
) -> Gtk.Widget:
    """Create widget for str-like plugin settings with multiline enabled."""
    wid = Gtk.ScrolledWindow()
    wid.set_shadow_type(type=Gtk.ShadowType.IN)  # pylint: disable=no-member
    wid.set_hexpand(True)
    # wid.set_vexpand(True)
    wid.set_size_request(50, 75)
    tview = Gtk.TextView()
    tview.set_border_width(6)  # pylint: disable=no-member
    buf = tview.get_buffer()
    buf.set_text(plugin_settings[setting])

    # we must use non-plain callback to get content of TextView.

    def callback(widget: Gtk.TextBuffer) -> None:
        start, end = buf.get_bounds()
        value = buf.get_text(start, end, True)
        setctl = settings.get_settings_controller()
        setctl.set_plugin_config(plugin_id, setting, value, str)

    buf.connect("changed", callback)
    wid.add(tview)  # pylint: disable=no-member
    return wid


def _make_plugin_sett_widget_bool(
    label: str,
    plugin_id: str,
    setting: str,
    plugin_settings: plugin_support.PluginSettings,
) -> Gtk.Widget:
    """Create combobox widget for plugin settings with type `bool`."""
    wid = Gtk.CheckButton.new_with_label(label)
    wid.set_active(plugin_settings[setting])
    wid.connect(
        "toggled",
        _create_plugin_change_cb(plugin_id, setting, bool, "get_active"),
    )
    return wid


def _make_plugin_sett_widget_int(
    plugin_id: str,
    setting: str,
    plugin_settings: plugin_support.PluginSettings,
) -> Gtk.Widget:
    """Make spinner button widget for plugin settings with type `int`."""
    wid = Gtk.SpinButton()
    wid.set_increments(1, 1)
    min_val, max_val = 0, 1000

    if val := plugin_settings.get_parameter(setting, "min"):
        min_val = int(val)

    if val := plugin_settings.get_parameter(setting, "max"):
        max_val = int(val)

    assert max_val > min_val

    wid.set_range(min_val, max_val)
    wid.set_value(plugin_settings[setting])
    wid.set_vexpand(False)

    wid.connect(
        "changed",
        _create_plugin_change_cb(
            plugin_id,
            setting,
            int,
            "get_text",
            no_false_values=True,
        ),
    )
    return wid


def _make_plugin_sett_widget_list(
    plugin_id: str,
    setting: str,
    plugin_settings: plugin_support.PluginSettings,
) -> Gtk.Widget | None:
    """Create widget for list-like plugin settings."""
    # TODO: plain (text) multiline

    helper = plugin_settings.get_parameter(setting, "helper")
    if helper == "choose_directory":
        return widgets.DirsSelectWidget(plugin_id, setting, plugin_settings)

    pretty.print_error("unknown helper", helper)
    return None


def _is_good_keystr(keystr: str) -> bool:
    """Check is `keystr` is acceptable key for global keybindings."""
    # Reject single letters so you can't bind 'A' etc
    if keystr is None:
        return False

    if label := Gtk.accelerator_get_label(*Gtk.accelerator_parse(keystr)):
        return not (len(label) == 1 and label.isalnum())

    return False


def _create_conf_keys_list() -> tuple[Gtk.TreeView, Gtk.ListStore]:
    columns = (_("Command"), _("Shortcut"), None)
    column_types = (str, str, str)

    keybind_store = Gtk.ListStore.new(column_types)
    keybind_table = Gtk.TreeView.new_with_model(keybind_store)
    for idx, col_header in enumerate(columns):
        renderer = Gtk.CellRendererText()
        column = Gtk.TreeViewColumn(col_header, renderer, text=idx)
        column.set_visible(col_header is not None)
        keybind_table.append_column(column)

    keybind_table.set_property("enable-search", False)
    keybind_table.set_headers_visible(True)
    keybind_table.show()
    return keybind_table, keybind_store


# pylint: disable=too-many-instance-attributes,too-many-public-methods
class PreferencesWindowController(pretty.OutputMixin):
    _instance: PreferencesWindowController | None = None
    _col_plugin_id = 0
    _col_enabled = 1
    _col_icon_name = 2
    _col_text = 3

    @classmethod
    def instance(cls) -> PreferencesWindowController:
        if cls._instance is None:
            cls._instance = PreferencesWindowController()

        return cls._instance

    def __init__(self):
        """Load ui from data file"""
        builder = Gtk.Builder()
        builder.set_translation_domain(version.PACKAGE_NAME)
        self.window: Gtk.Window = None

        if ui_file := config.get_data_file("preferences.ui"):
            builder.add_from_file(ui_file)
        else:
            return

        self.window = builder.get_object("preferenceswindow")
        self.window.set_position(Gtk.WindowPosition.CENTER)
        self.window.connect("delete-event", self._on_wnd_delete_event)
        self._plugin_sidebar = builder.get_object("plugin_about_parent")
        self._preferences_notebook = builder.get_object("preferences_notebook")
        self._btn_remove_dir = builder.get_object("buttonremovedirectory")
        self._plugins_filter = builder.get_object("entry_plugins_filter")
        self._sources_list_ctrl = _SourceListController(
            builder.get_object("source_list_parent")
        )

        builder.get_object("checkautostart").set_active(
            self._is_autostart_enabled()
        )

        setctl = settings.get_settings_controller()
        _set_combobox(
            setctl.get_config_int("Appearance", "icon_large_size"),
            builder.get_object("icons_large_size"),
        )
        _set_combobox(
            setctl.get_config_int("Appearance", "icon_small_size"),
            builder.get_object("icons_small_size"),
        )
        builder.get_object("checkusecommandkeys").set_active(
            setctl.get_use_command_keys()
        )
        builder.get_object("radio_actionaccelalt").set_active(
            setctl.get_action_accelerator_modifer() != "ctrl"
        )
        builder.get_object("radio_actionaccelctrl").set_active(
            setctl.get_action_accelerator_modifer() == "ctrl"
        )

        self._init_checkstatus(setctl, builder)

        # Make alternative comboboxes
        self._terminal_combobox = builder.get_object("terminal_combobox")
        _make_combobox_model(self._terminal_combobox)
        self._update_alternative_combobox("terminal", self._terminal_combobox)

        self._editor_combobox = builder.get_object("editor_combobox")
        _make_combobox_model(self._editor_combobox)
        self._update_alternative_combobox("editor", self._editor_combobox)

        self._icons_combobox = builder.get_object("icons_combobox")
        _make_combobox_model(self._icons_combobox)
        self._update_alternative_combobox(
            "icon_renderer", self._icons_combobox
        )

        _set_combobox_id(
            setctl.get_config_int("Appearance", "ellipsize_mode"),
            builder.get_object("ellipsize_mode"),
        )

        # Plugin List
        self._init_plugin_lists(builder.get_object("plugin_list_parent"))

        # Directory List
        self._init_indexed_folders_widgets(
            builder.get_object("directory_list_parent")
        )
        self._load_indexed_folders_settings()

        # global keybindings list
        self._keybind_table, self._keybind_store = _create_conf_keys_list()
        builder.get_object("keybindings_list_parent").add(self._keybind_table)
        self._keybind_table.connect(
            "row-activated", self.on_keybindings_row_activate
        )
        builder.get_object("button_reset_keys").set_sensitive(
            keybindings.is_available()
        )
        self._keybind_table.set_sensitive(keybindings.is_available())

        # kupfer interface (accelerators) keybindings list
        self._init_keybindings(builder.get_object("gkeybindings_list_parent"))
        self._show_keybindings(setctl)
        self._show_gkeybindings(setctl)

        # Connect to signals at the last point
        builder.connect_signals(self)  # pylint: disable=no-member
        setctl.connect("alternatives-changed", self._on_alternatives_changed)

    def _init_checkstatus(
        self, setctl: settings.SettingsController, builder: Gtk.Builder
    ) -> None:
        checkstatusicon_gtk = builder.get_object("checkstatusicon_gtk")
        checkstatusicon_gtk.set_label(
            checkstatusicon_gtk.get_label() + " (GtkStatusIcon)"
        )
        checkstatusicon_gtk.set_active(setctl.get_show_status_icon())

        checkstatusicon_ai = builder.get_object("checkstatusicon_ai")
        checkstatusicon_ai.set_label(
            checkstatusicon_ai.get_label() + " (AppIndicator)"
        )
        if _supports_app_indicator():
            checkstatusicon_ai.set_active(setctl.get_show_status_icon_ai())
        else:
            checkstatusicon_ai.set_sensitive(False)

    def _init_plugin_lists(self, parent: Gtk.Widget) -> None:
        # setup plugin list table
        # cols: ("plugin_id", "enabled", "icon-name", "text")
        self._plugins_store = Gtk.ListStore.new((str, bool, str, str))
        self._plugins_table = table = Gtk.TreeView.new_with_model(
            self._plugins_store
        )
        table.set_headers_visible(False)
        table.set_property("enable-search", False)
        table.connect("cursor-changed", self._on_plugins_sel_changed)
        table.get_selection().set_mode(Gtk.SelectionMode.BROWSE)

        checkcell = Gtk.CellRendererToggle()
        checkcol = Gtk.TreeViewColumn("item", checkcell)
        checkcol.add_attribute(checkcell, "active", self._col_enabled)
        checkcell.connect("toggled", self._on_plugin_enable_toggled)

        icon_cell = Gtk.CellRendererPixbuf()
        icon_cell.set_property("height", _LIST_ICON_SIZE)
        icon_cell.set_property("width", _LIST_ICON_SIZE)

        icon_col = Gtk.TreeViewColumn("icon", icon_cell)
        icon_col.add_attribute(icon_cell, "icon-name", self._col_icon_name)

        cell = Gtk.CellRendererText()
        col = Gtk.TreeViewColumn("item", cell)
        col.add_attribute(cell, "text", self._col_text)

        table.append_column(checkcol)
        # hide icon for now
        # self.table.append_column(icon_col)
        table.append_column(col)

        self.plugin_list_timer = scheduler.Timer()
        self.plugin_info = kupferstring.locale_sort(
            plugins.get_plugin_info(),
            key=operator.itemgetter("localized_name"),
        )

        self._refresh_plugin_list()
        self.output_debug(f"Standard Plugins: {len(self._plugins_store)}")
        table.show()

        parent.add(self._plugins_table)

    def _init_indexed_folders_widgets(self, parent: Gtk.Widget) -> None:
        """Setup indexed folders list & model."""
        self._indexed_folders_store = Gtk.ListStore.new([str, Gio.Icon, str])
        self._indexed_folders_table = table = Gtk.TreeView.new_with_model(
            self._indexed_folders_store
        )
        table.set_headers_visible(False)
        table.set_property("enable-search", False)
        table.connect("cursor-changed", self._on_indexed_folders_sel_changed)
        table.get_selection().set_mode(Gtk.SelectionMode.BROWSE)

        icon_cell = Gtk.CellRendererPixbuf()
        icon_col = Gtk.TreeViewColumn("icon", icon_cell)
        icon_col.add_attribute(icon_cell, "gicon", 1)

        cell = Gtk.CellRendererText()
        cell.set_property("ellipsize", Pango.EllipsizeMode.END)

        col = Gtk.TreeViewColumn("name", cell)
        col.add_attribute(cell, "text", 2)

        table.append_column(icon_col)
        table.append_column(col)
        table.show()

        parent.add(self._indexed_folders_table)

    def _init_keybindings(self, parent: Gtk.Widget) -> None:
        self.gkeybind_table, self.gkeybind_store = _create_conf_keys_list()
        parent.add(self.gkeybind_table)
        self.gkeybind_table.connect(
            "row-activated", self.on_gkeybindings_row_activate
        )

        # Requires GTK 3.22
        with suppress(AttributeError):
            parent.set_propagate_natural_height(True)

    def _show_keybindings(self, setctl: settings.SettingsController) -> None:
        names = (
            # TRANS: Names of global keyboard shortcuts
            (_("Show Main Interface"), "keybinding"),
            (_("Show with Selection"), "magickeybinding"),
        )
        self._keybind_store.clear()
        for name, binding in sorted(names):
            accel = setctl.get_global_keybinding(binding)
            label = Gtk.accelerator_get_label(*Gtk.accelerator_parse(accel))
            self._keybind_store.append((name, label, binding))

    def _show_gkeybindings(self, setctl: settings.SettingsController) -> None:
        names = accelerators.ACCELERATOR_NAMES
        self.gkeybind_store.clear()
        for binding in sorted(names, key=lambda k: names[k]):
            accel = setctl.get_accelerator(binding) or ""
            label = Gtk.accelerator_get_label(*Gtk.accelerator_parse(accel))
            self.gkeybind_store.append((names[binding], label, binding))

    def _load_indexed_folders_settings(self) -> None:
        """Load indexed folders."""
        setctl = settings.get_settings_controller()
        for directory in setctl.get_directories():
            self._add_indexed_directory(directory, store=False)

    def _add_indexed_directory(
        self, directory: str, store: bool = False
    ) -> None:
        """Add directory to list indexed folders. When `store` update list in
        SettingController."""
        have = [
            os.path.normpath(row[0]) for row in self._indexed_folders_store
        ]
        if directory in have:
            self.output_debug("Ignoring duplicate directory: ", directory)
            return

        have.append(directory)
        directory = os.path.expanduser(directory)
        dispname = launch.get_display_path_for_bytestring(directory)
        gicon = icons.get_gicon_for_file(directory)
        self._indexed_folders_store.append((directory, gicon, dispname))

        if store:
            setctl = settings.get_settings_controller()
            setctl.set_directories(have)

    def _remove_indexed_folder(
        self, rowiter: Gtk.TreeIter, store: bool = True
    ) -> None:
        """Remove directory from list indexed folders by `rowiter`. When `store`
        update list in SettingController."""
        self._indexed_folders_store.remove(rowiter)
        if store:
            have = [
                os.path.normpath(row[0]) for row in self._indexed_folders_store
            ]
            setctl = settings.get_settings_controller()
            setctl.set_directories(have)

    def on_preferenceswindow_key_press_event(
        self, widget: Gtk.Widget, event: Gdk.EventKey
    ) -> bool:
        if event.keyval == Gdk.keyval_from_name("Escape"):
            self._hide()
            return True

        return False

    def on_checkstatusicon_gtk_toggled(self, widget: Gtk.Widget) -> None:
        """Toggle status icon checkbox callback."""
        setctl = settings.get_settings_controller()
        setctl.set_show_status_icon(widget.get_active())

    def on_checkstatusicon_ai_toggled(self, widget: Gtk.Widget) -> None:
        """Toggle AppIndicator3 status icon checkbox callback."""
        setctl = settings.get_settings_controller()
        setctl.set_show_status_icon_ai(widget.get_active())

    def _is_autostart_enabled(self) -> bool:
        """Get state of autostart settings."""
        autostart_dir = xdgBase.save_config_path("autostart")
        autostart_file = Path(autostart_dir, _KUPFER_DESKTOP)
        if not autostart_file.exists():
            return False

        try:
            dfile = DesktopEntry.DesktopEntry(str(autostart_file))
        except xdgExc.ParsingError as exception:
            pretty.print_error(__name__, exception)
            return False

        return (
            dfile.hasKey(_AUTOSTART_KEY)
            and dfile.get(_AUTOSTART_KEY, type="boolean")
        ) and (
            not dfile.hasKey(_HIDDEN_KEY)
            or not dfile.get(_HIDDEN_KEY, type="boolean")
        )

    def on_checkautostart_toggled(self, widget: Gtk.Widget) -> None:
        """Change autostart settings on toggle autostart checkbox."""
        autostart_dir = xdgBase.save_config_path("autostart")
        autostart_file = Path(autostart_dir, _KUPFER_DESKTOP)
        if not autostart_file.exists():
            desktop_files = list(
                xdgBase.load_data_paths("applications", _KUPFER_DESKTOP)
            )
            if not desktop_files:
                self.output_error("Installed kupfer desktop file not found!")
                return

            desktop_file_path = desktop_files[0]
            # Read installed file and modify it
            try:
                dfile = DesktopEntry.DesktopEntry(desktop_file_path)
            except xdgExc.ParsingError as exception:
                pretty.print_error(__name__, exception)
                return

            executable = dfile.getExec()
            ## append no-splash
            if "--no-splash" not in executable:
                executable += " --no-splash"

            dfile.set("Exec", executable)
        else:
            try:
                dfile = DesktopEntry.DesktopEntry(str(autostart_file))
            except xdgExc.ParsingError as exception:
                pretty.print_error(__name__, exception)
                return

        activestr = str(bool(widget.get_active())).lower()
        not_activestr = str(not bool(widget.get_active())).lower()
        self.output_debug("Setting autostart to", activestr)
        dfile.set(_AUTOSTART_KEY, activestr)
        dfile.set(_HIDDEN_KEY, not_activestr)
        ## remove the format specifiers
        executable = dfile.getExec().replace("%F", "")
        dfile.set("Exec", executable)
        dfile.write(filename=str(autostart_file))

    def on_helpbutton_clicked(self, widget: Gtk.Widget) -> None:
        kupferhelp.show_help()

    def on_closebutton_clicked(self, widget: Gtk.Widget) -> None:
        self._hide()

    def _refresh_plugin_list(self, us_filter: str | None = None) -> None:
        """List plugins that pass text filter @us_filter or list all if None"""
        self._plugins_store.clear()
        setctl = settings.get_settings_controller()

        if us_filter:
            self.plugin_list_timer.set_ms(300, self._show_focus_topmost_plugin)
        else:
            self.plugin_list_timer.invalidate()

        for info in self.plugin_info:
            plugin_id = info["name"]
            if setctl.get_plugin_is_hidden(plugin_id):
                continue

            name = info["localized_name"]

            if us_filter:
                name_score = relevance.score(name, us_filter)
                folded_name = kupferstring.tofolded(name)
                fold_name_score = relevance.score(folded_name, us_filter)
                desc_score = relevance.score(info["description"], us_filter)
                if not name_score and not fold_name_score and desc_score < 0.9:  # noqa:PLR2004
                    continue

            enabled = setctl.get_plugin_enabled(plugin_id)
            self._plugins_store.append(
                (plugin_id, enabled, "kupfer-object", str(name))
            )

    def _show_focus_topmost_plugin(self) -> None:
        try:
            first_row = next(iter(self._plugins_store))
        except StopIteration:
            return

        plugin_id = first_row[0]
        self.show_focus_plugin(plugin_id, 0)

    def _on_plugin_enable_toggled(
        self, cell: Gtk.CellRendererToggle, path: str
    ) -> None:
        plugin_id = self._get_plugin_id(path)
        pathit = self._plugins_store.get_iter(path)
        plugin_is_enabled = not self._plugins_store.get_value(
            pathit, self._col_enabled
        )
        self._plugins_store.set_value(
            pathit, self._col_enabled, plugin_is_enabled
        )
        setctl = settings.get_settings_controller()
        setctl.set_plugin_enabled(plugin_id, plugin_is_enabled)
        self._plugin_sidebar_update(plugin_id)

    def _get_plugin_id(self, path: str | Gtk.TreePath) -> str:
        """Find plugin id by tree `path`."""
        pathit = self._plugins_store.get_iter(path)
        plugin_id: str = self._plugins_store.get_value(
            pathit, self._col_plugin_id
        )
        return plugin_id

    def _table_path_for_plugin_id(self, plugin_id: str) -> Gtk.TreePath:
        """Find the tree path of @plugin_id"""
        for row in self._plugins_store:
            if plugin_id == row[self._col_plugin_id]:
                return row.path

        raise ValueError(f"No such plugin {plugin_id}")

    def _get_plugin_info(self, plugin_id: str) -> dict[str, ty.Any] | None:
        """Get dict with plugin information for `plugin_id`."""
        for info in self.plugin_info:
            if info["name"] == plugin_id:
                return info

        return None

    def _on_plugins_sel_changed(self, table: Gtk.TreeView) -> None:
        curpath, _curcol = table.get_cursor()
        if not curpath:
            return

        plugin_id = self._get_plugin_id(curpath)
        self._plugin_sidebar_update(plugin_id)

    def _plugin_sidebar_update(self, plugin_id: str) -> None:
        """Update content of plugin sidebar panel.
        Remove old panel if exist. Create new widgets with information and
        settings.
        """
        # remove old panel if exists
        if oldch := self._plugin_sidebar.get_child():
            self._plugin_sidebar.remove(oldch)
            oldch.destroy()

        info = self._get_plugin_info(plugin_id)
        if not info:
            return

        about = Gtk.Box.new(Gtk.Orientation.VERTICAL, 12)

        # title
        label = Gtk.Label()
        label.set_alignment(0, 0)  # pylint: disable=no-member
        title = GLib.markup_escape_text(info["localized_name"])
        label.set_markup(f"<b><big>{title}</big></b>")
        label.set_line_wrap(True)  # pylint: disable=no-member
        label.set_selectable(True)
        about.pack_start(label, False, True, 0)

        # about section
        about_widg = widgets.PluginAboutWidget(plugin_id, info)
        about.pack_start(about_widg, False, True, 0)

        # objects info section
        if objects := self._make_plugin_objects_widget(plugin_id):
            about.pack_start(objects, False, True, 0)

        # settings
        if psettings_wid := self._make_plugin_settings_widget(plugin_id):
            about.pack_start(psettings_wid, True, True, 0)

        # create new panel
        vport = Gtk.Viewport()
        vport.set_shadow_type(Gtk.ShadowType.NONE)  # pylint: disable=no-member
        vport.add(about)  # pylint: disable=no-member
        self._plugin_sidebar.add(vport)
        self._plugin_sidebar.show_all()

    def _make_plugin_objects_widget(self, plugin_id: str) -> Gtk.Widget | None:
        srcs, actions, text_sources = plugins.get_plugin_attributes(
            plugin_id,
            (
                plugins.PluginAttr.SOURCES,
                plugins.PluginAttr.ACTION_DECORATORS,
                plugins.PluginAttr.TEXT_SOURCES,
            ),
        )

        if not actions and not srcs and not text_sources:
            return None

        vbox = Gtk.Box.new(Gtk.Orientation.VERTICAL, 6)

        if srcs := list(srcs or ()) + list(text_sources or ()):
            # TRANS: Plugin contents header
            widgets.new_label_header(vbox, _("Sources"))
            swid = widgets.ObjectsInfoWidget(plugin_id, srcs)
            vbox.pack_start(swid, True, True, 0)

        if actions:  # pylint: disable=using-constant-test
            # TRANS: Plugin contents header
            widgets.new_label_header(vbox, _("Actions"))
            awid = widgets.ObjectsInfoWidget(plugin_id, actions)
            vbox.pack_start(awid, True, True, 0)

        return vbox

    # pylint: disable=too-many-locals
    def _make_plugin_settings_widget(
        self, plugin_id: str
    ) -> Gtk.Widget | None:
        plugin_settings: plugin_support.PluginSettings | None
        plugin_settings = plugins.get_plugin_attribute(
            plugin_id, plugins.PluginAttr.SETTINGS
        )
        if not plugin_settings:
            return None

        box = Gtk.Grid()
        box.set_row_spacing(6)
        box.set_column_spacing(12)
        label = widgets.new_label_header(None, _("Configuration"))
        box.attach(label, 0, 0, 2, 1)

        for row, setting in enumerate(plugin_settings, 1):
            label = plugin_settings.get_label(setting)
            typ = plugin_settings.get_value_type(setting)

            tooltip = plugin_settings.get_tooltip(setting)
            if issubclass(typ, plugin_support.UserNamePassword):
                wid = Gtk.Button(label or _("Set username and password"))
                wid.set_tooltip_text(tooltip)
                wid.connect(
                    "clicked",
                    _create_plugin_credentials_cb(plugin_id, setting),
                )
                box.attach(wid, 0, row, 2, 1)
                continue

            if issubclass(typ, bool):
                wid = _make_plugin_sett_widget_bool(
                    label, plugin_id, setting, plugin_settings
                )
                wid.set_tooltip_text(tooltip)
                box.attach(wid, 0, row, 2, 1)
                continue

            wid = None
            multiline = False
            if issubclass(typ, str):
                wid, multiline = _make_plugin_sett_widget_str(
                    plugin_id, setting, plugin_settings
                )
                box.attach(wid, 1, row, 1, 1)

            elif issubclass(typ, int):
                wid = _make_plugin_sett_widget_int(
                    plugin_id, setting, plugin_settings
                )
                box.attach(wid, 1, row, 1, 1)

            elif issubclass(typ, list):
                if wid := _make_plugin_sett_widget_list(
                    plugin_id, setting, plugin_settings
                ):
                    multiline = True
                    box.attach(wid, 1, row, 1, 1)

            wid.set_tooltip_text(tooltip)

            label_wid = widgets.new_label(
                None, label, selectable=False, tooltip=tooltip
            )
            box.attach(label_wid, 0, row, 1, 1)
            # on multiline widgets put label on top
            if multiline:
                label_wid.set_alignment(0, 0)  # pylint: disable=no-member
                label_wid.set_padding(0, 4)  # pylint: disable=no-member
            else:
                label_wid.set_alignment(0, 0.5)  # pylint: disable=no-member
                label_wid.set_padding(0, 0)  # pylint: disable=no-member

        box.show_all()  # pylint: disable=no-member
        return box

    def on_buttonadddirectory_clicked(self, widget: Gtk.Widget) -> None:
        """Choose folder for indexed folders - callback."""
        # TRANS: File Chooser Title
        chooser_dialog = Gtk.FileChooserDialog(
            title=_("Choose a Directory"),
            action=Gtk.FileChooserAction.SELECT_FOLDER,
            buttons=(
                Gtk.STOCK_CANCEL,
                Gtk.ResponseType.REJECT,
                Gtk.STOCK_OK,
                Gtk.ResponseType.ACCEPT,
            ),
        )
        chooser_dialog.set_select_multiple(True)
        # pylint: disable=no-member
        if chooser_dialog.run() == Gtk.ResponseType.ACCEPT:
            # pylint: disable=no-member
            for selected_dir in chooser_dialog.get_filenames():
                self._add_indexed_directory(selected_dir, store=True)

        chooser_dialog.hide()

    def on_buttonremovedirectory_clicked(self, widget: Gtk.Widget) -> None:
        """Fromove folder from indexed folders - callback."""
        curpath, _curcol = self._indexed_folders_table.get_cursor()
        if curpath:
            pathit = self._indexed_folders_store.get_iter(curpath)
            self._remove_indexed_folder(pathit, store=True)

    def on_entry_plugins_filter_changed(self, widget: Gtk.Widget) -> None:
        """Filter plugins list on filter entry changed - callback."""
        s_filter = widget.get_text()
        us_filter = s_filter.lower()
        self._refresh_plugin_list(us_filter)

    def on_entry_plugins_filter_icon_press(
        self, entry: Gtk.Entry, icon_pos: ty.Any, event: ty.Any
    ) -> None:
        """Clear plugins filter, refresh list - callback."""
        entry.set_text("")

    def on_keybindings_row_activate(
        self,
        treeview: Gtk.TreeView,
        path: Gtk.TreePath,
        view_column: Gtk.TreeViewColumn,
    ) -> None:
        """Change keybinding (ask user for new keybinding) on activate row
        of keybindings list - callback."""

        def bind_key_func(target):
            def bind_key(keystr):
                return keybindings.bind_key(keystr, target)

            return bind_key

        pathit = self._keybind_store.get_iter(path)
        keybind_id = self._keybind_store.get_value(pathit, 2)
        setctl = settings.get_settings_controller()
        curr_key = setctl.get_global_keybinding(keybind_id)
        bind_func = bind_key_func(_KEYBINDING_TARGETS[keybind_id])
        keystr = getkey_dialog.ask_for_key(
            bind_func,
            curr_key,
            screen=treeview.get_screen(),
            parent=treeview.get_toplevel(),
        )
        if keystr == "":
            keybindings.bind_key(None, _KEYBINDING_TARGETS[keybind_id])
            setctl.set_global_keybinding(keybind_id, keystr)
            self._keybind_store.set_value(pathit, 1, "")

        elif keystr is not None:
            setctl.set_global_keybinding(keybind_id, keystr)
            label = Gtk.accelerator_get_label(*Gtk.accelerator_parse(keystr))
            self._keybind_store.set_value(pathit, 1, label)

    def on_gkeybindings_row_activate(
        self,
        treeview: Gtk.TreeView,
        path: Gtk.TreePath,
        view_column: Gtk.TreeViewColumn,
    ) -> None:
        """Ask user for new global keybinding on activate row of global
        keybindings list."""
        pathit = self.gkeybind_store.get_iter(path)
        keybind_id = self.gkeybind_store.get_value(pathit, 2)
        setctl = settings.get_settings_controller()
        curr_key = setctl.get_accelerator(keybind_id)
        keystr = getkey_dialog.ask_for_key(
            _is_good_keystr,
            previous_key=curr_key,
            screen=treeview.get_screen(),
            parent=treeview.get_toplevel(),
        )

        if keystr is not None:
            setctl.set_accelerator(keybind_id, keystr)
            label = Gtk.accelerator_get_label(*Gtk.accelerator_parse(keystr))
            self.gkeybind_store.set_value(pathit, 1, label)

    def on_button_reset_keys_clicked(self, button: Gtk.Button) -> None:
        """Reset keybindings to defaults after confirmation."""
        if self._ask_user_for_reset_keybinding():
            setctl = settings.get_settings_controller()
            setctl.reset_keybindings()
            self._show_keybindings(setctl)
            # Unbind all before re-binding
            for target in _KEYBINDING_TARGETS.values():
                keybindings.bind_key(None, target)

            for keybind_id, target in _KEYBINDING_TARGETS.items():
                keystr = setctl.get_global_keybinding(keybind_id)
                keybindings.bind_key(keystr, target)

    def on_button_reset_gkeys_clicked(self, button: Gtk.Button) -> None:
        """Reset global keybindings to defaults after confirmation."""
        if self._ask_user_for_reset_keybinding():
            setctl = settings.get_settings_controller()
            setctl.reset_accelerators()
            self._show_gkeybindings(setctl)

    def on_checkusecommandkeys_toggled(self, widget: Gtk.Widget) -> None:
        """Change 'use single keystroke commands...' setting - callback."""
        setctl = settings.get_settings_controller()
        setctl.set_use_command_keys(widget.get_active())

    def on_radio_action_accel_alt(self, widget: Gtk.RadioButton) -> None:
        """Change 'Action accelerators use alt' setting - callback."""
        if widget.get_active():
            setctl = settings.get_settings_controller()
            setctl.set_action_accelerator_modifier("alt")

    def on_radio_action_accel_ctrl(self, widget: Gtk.RadioButton) -> None:
        """Change 'Action accelerators use ctrl' setting - callback."""
        if widget.get_active():
            setctl = settings.get_settings_controller()
            setctl.set_action_accelerator_modifier("ctrl")

    def _on_indexed_folders_sel_changed(self, table: Gtk.TreeView) -> None:
        curpath, _curcol = table.get_cursor()
        if not curpath or not self._indexed_folders_store:
            self._btn_remove_dir.set_sensitive(False)
            return

        self._btn_remove_dir.set_sensitive(True)

    def on_terminal_combobox_changed(self, widget: Gtk.ComboBox) -> None:
        """Change terminal setting - callback."""
        setctl = settings.get_settings_controller()
        if itr := widget.get_active_iter():
            term_id = widget.get_model().get_value(itr, 1)
            setctl.set_preferred_tool("terminal", term_id)

    def on_editor_combobox_changed(self, widget: Gtk.ComboBox) -> None:
        """Change editor setting - callback."""
        setctl = settings.get_settings_controller()
        if itr := widget.get_active_iter():
            editor_id = widget.get_model().get_value(itr, 1)
            setctl.set_preferred_tool("editor", editor_id)

    def on_icons_combobox_changed(self, widget: Gtk.ComboBox) -> None:
        """Change 'icons set' setting - callback."""
        setctl = settings.get_settings_controller()
        if itr := widget.get_active_iter():
            term_id = widget.get_model().get_value(itr, 1)
            setctl.set_preferred_tool("icon_renderer", term_id)

    def on_icons_large_size_changed(self, widget: Gtk.ComboBoxText) -> None:
        """Change 'large icon size' setting - callback."""
        if widget.get_active_iter():
            val = widget.get_active_text()
            setctl = settings.get_settings_controller()
            setctl.set_large_icon_size(val)

    def on_icons_small_size_changed(self, widget: Gtk.ComboBoxText) -> None:
        """Change 'small icon size' setting - callback."""
        if widget.get_active_iter():
            val = widget.get_active_text()
            setctl = settings.get_settings_controller()
            setctl.set_small_icon_size(val)

    def on_ellipsize_mode_changed(self, widget: Gtk.ComboBox) -> None:
        """Change 'Text ellipsization' setting - callback."""
        setctl = settings.get_settings_controller()
        if itr := widget.get_active_iter():
            mode = widget.get_model().get_value(itr, 1)
            setctl.set_config("Appearance", "ellipsize_mode", mode)

    def _update_alternative_combobox(
        self, category_key: str, combobox: Gtk.ComboBox
    ) -> None:
        """
        Alternatives changed
        """
        combobox_store = combobox.get_model()
        combobox_store.clear()
        setctl = settings.get_settings_controller()
        term_id = setctl.get_preferred_tool(category_key)
        # fill in the available alternatives
        alternatives = kupferstring.locale_sort(
            setctl.get_valid_alternative_ids(category_key),
            key=operator.itemgetter(1),
        )
        term_iter = None
        for id_, name in alternatives:
            _it = combobox_store.append((name, id_))
            if id_ == term_id:
                term_iter = _it

        # Update selection
        term_iter = term_iter or combobox_store.get_iter_first()
        combobox.set_sensitive(len(combobox_store) > 1)
        if term_iter:
            combobox.set_active_iter(term_iter)

    def _on_alternatives_changed(
        self, setctl: settings.SettingsController, category_key: str
    ) -> None:
        """Update alternatives comboboxes after settings changed."""
        if category_key == "terminal":
            self._update_alternative_combobox(
                category_key, self._terminal_combobox
            )

        elif category_key == "editor":
            self._update_alternative_combobox(
                category_key, self._editor_combobox
            )

        elif category_key == "icon_renderer":
            self._update_alternative_combobox(
                category_key, self._icons_combobox
            )

    def on_preferences_notebook_switch_page(
        self, notebook: Gtk.Notebook, page: Gtk.Widget, page_num: int
    ) -> None:
        ## focus the search box on the plugin tab
        if page_num == _PLUGIN_LIST_PAGE:
            GLib.idle_add(self._plugins_filter.grab_focus)

    def show(self, timestamp: int) -> None:
        assert self.window
        self.window.present_with_time(timestamp)

    def show_on_screen(self, timestamp: int, screen: Gdk.Screen) -> None:
        assert self.window
        self.window.set_screen(screen)
        self.show(timestamp)
        ## focus the search box on the plugin tab
        if self._preferences_notebook.get_current_page() == _PLUGIN_LIST_PAGE:
            self._plugins_filter.grab_focus()

    def show_focus_plugin(self, plugin_id: str, timestamp: int) -> None:
        """Open and show information about plugin `plugin_id`."""
        assert self.window

        try:
            table_path = self._table_path_for_plugin_id(plugin_id)
        except ValueError:
            self._plugins_filter.set_text("")
            self._refresh_plugin_list()
            table_path = self._table_path_for_plugin_id(plugin_id)

        self._plugins_table.set_cursor(table_path)
        self._plugins_table.scroll_to_cell(table_path)
        self._preferences_notebook.set_current_page(_PLUGIN_LIST_PAGE)
        self.window.present_with_time(timestamp)

    def _hide(self) -> None:
        assert self.window
        # close window instead of hide
        # self.window.hide()
        self.window.close()

    def _on_wnd_delete_event(self, *ignored: ty.Any) -> bool:
        # remove instance on close window
        # self._hide()
        self.window.destroy()
        PreferencesWindowController._instance = None
        return False

    def _ask_user_for_reset_keybinding(self) -> bool:
        """Request user confirmation of resetting keybinding."""
        dlg = Gtk.MessageDialog(
            self.window, Gtk.DialogFlags.MODAL, Gtk.MessageType.QUESTION
        )
        dlg.set_markup(_("Reset all shortcuts to default values?"))
        dlg.add_buttons(
            Gtk.STOCK_CANCEL,
            Gtk.ResponseType.CLOSE,
            _("Reset"),
            Gtk.ResponseType.ACCEPT,
        )
        # pylint: disable=no-member
        result: bool = dlg.run() == Gtk.ResponseType.ACCEPT
        dlg.destroy()
        return result


get_preferences_window_controller = PreferencesWindowController.instance


# pylint: disable=too-few-public-methods
class _SourceListController:
    """Controller for sources list."""

    def __init__(self, parent_widget):
        # setup plugin list table
        column_types = (GObject.TYPE_PYOBJECT, str, bool, Gio.Icon, str)
        columns = ("source", "plugin_id", "toplevel", "icon", "text")
        self.store = Gtk.ListStore(*column_types)
        self.table = Gtk.TreeView.new_with_model(self.store)
        self.table.set_headers_visible(False)
        self.table.set_property("enable-search", False)
        # self.table.connect("cursor-changed", self.plugin_table_cursor_changed)
        self.table.get_selection().set_mode(Gtk.SelectionMode.NONE)

        checkcell = Gtk.CellRendererToggle()
        checkcol = Gtk.TreeViewColumn("item", checkcell)
        checkcol.add_attribute(checkcell, "active", columns.index("toplevel"))
        checkcell.connect("toggled", self.on_checktoplevel_enabled)

        icon_cell = Gtk.CellRendererPixbuf()
        icon_cell.set_property("height", _LIST_ICON_SIZE)
        icon_cell.set_property("width", _LIST_ICON_SIZE)

        icon_col = Gtk.TreeViewColumn("icon", icon_cell)
        icon_col.add_attribute(icon_cell, "gicon", columns.index("icon"))

        cell = Gtk.CellRendererText()
        col = Gtk.TreeViewColumn("item", cell)
        col.add_attribute(cell, "text", columns.index("text"))

        self.table.append_column(checkcol)
        self.table.append_column(icon_col)
        self.table.append_column(col)

        self._refresh()
        self.table.show()
        parent_widget.add(self.table)

        setctl = settings.get_settings_controller()
        setctl.connect("plugin-enabled-changed", self._refresh)

    def _refresh(self, *ignored: ty.Any) -> None:
        self.store.clear()
        setctl = settings.get_settings_controller()
        srcctl = sources.get_source_controller()
        srcs = sorted(srcctl.get_sources(), key=str)

        for src in srcs:
            plugin_id = srcctl.get_plugin_id_for_object(src)
            if not plugin_id or setctl.get_plugin_is_hidden(plugin_id):
                continue

            if not src.get_valid_leaf_repr()[0]:
                continue

            gicon = src.get_icon()
            toplevel = setctl.get_source_is_toplevel(plugin_id, src)
            name = str(src)
            self.store.append((src, plugin_id, toplevel, gicon, name))

    def on_checktoplevel_enabled(
        self, cell: Gtk.CellRendererToggle, path: str
    ) -> None:
        pathit = self.store.get_iter(path)
        is_toplevel = not self.store.get_value(pathit, 2)
        plugin_id = self.store.get_value(pathit, 1)
        src = self.store.get_value(pathit, 0)

        srcctl = sources.get_source_controller()
        srcctl.set_toplevel(src, is_toplevel)

        setctl = settings.get_settings_controller()
        setctl.set_source_is_toplevel(plugin_id, src, is_toplevel)
        self.store.set_value(pathit, 2, is_toplevel)


def _supports_app_indicator() -> bool:
    try:
        gi.require_version("AyatanaAppIndicator3", "0.1")
    except ValueError:
        return False

    return True


def _get_time(ctxenv: GUIEnvironmentContext | None) -> int:
    if ctxenv:
        return ctxenv.get_timestamp()

    return int(Gtk.get_current_event_time())


def show_preferences(ctxenv: GUIEnvironmentContext) -> None:
    win = get_preferences_window_controller()
    if ctxenv:
        win.show_on_screen(ctxenv.get_timestamp(), ctxenv.get_screen())
    else:
        win.show(_get_time(ctxenv))


def show_plugin_info(
    plugin_id: str, ctxenv: GUIEnvironmentContext | None = None
) -> None:
    prefs = get_preferences_window_controller()
    prefs.show_focus_plugin(plugin_id, _get_time(ctxenv))



from __future__ import annotations

import typing as ty

import gi
from gi.repository import GObject

from kupfer import environment
from kupfer.support import pretty

Keybinder: ty.Any = None  # pylint: disable=invalid-name
if environment.allows_keybinder():
    try:
        gi.require_version("Keybinder", "3.0")

        # pylint: disable=ungrouped-imports
        from gi.repository import Keybinder  # type:ignore

        Keybinder.init()
    except (ValueError, ImportError):
        pretty.print_debug(__name__, "Keybinder 3.0 not available in gi")
        Keybinder = None

else:
    pretty.print_debug(__name__, "Keybinder disabled")

KeybindingTarget = int

KEYBINDING_TARGET_DEFAULT = 1
KEYBINDING_TARGET_MAGIC = 2
KEYRANGE_RESERVED = (3, 0x1000)
# range for custom keybingings
KEYRANGE_TRIGGERS = (0x1000, 0x2000)


def get_keybound_object():
    """Get the shared instance"""
    return KeyboundObject.instance()


class KeyboundObject(GObject.GObject):  # type:ignore
    """Keybinder object

    signals:
        keybinding (target, event_time)
        keybinding signal is triggered when the key bound
        for @target is triggered.
    """

    __gtype_name__ = "KeyboundObject"
    _instance: KeyboundObject | None = None

    @classmethod
    def instance(cls) -> KeyboundObject:
        if cls._instance is None:
            cls._instance = KeyboundObject()

        return cls._instance

    def keybinding(self, target: int) -> None:
        assert Keybinder
        time = Keybinder.get_current_event_time()
        self.emit("keybinding", target, "", time)

    def emit_bound_key_changed(self, keystring: str, is_bound: bool) -> None:
        self.emit("bound-key-changed", keystring, is_bound)

    def relayed_keys(
        self,
        _sender: ty.Any,
        keystring: str,
        display: str,
        timestamp: float,
    ) -> None:
        for target, key in _CURRENTLY_BOUND.items():
            if keystring == key:
                self.emit("keybinding", target, display, timestamp)


# Arguments: Target, Display, Timestamp
GObject.signal_new(
    "keybinding",
    KeyboundObject,
    GObject.SignalFlags.RUN_LAST,
    GObject.TYPE_BOOLEAN,
    (GObject.TYPE_INT, GObject.TYPE_STRING, GObject.TYPE_UINT),
)
# Arguments: Keystring, Boolean
GObject.signal_new(
    "bound-key-changed",
    KeyboundObject,
    GObject.SignalFlags.RUN_LAST,
    GObject.TYPE_BOOLEAN,
    (
        GObject.TYPE_STRING,
        GObject.TYPE_BOOLEAN,
    ),
)

_CURRENTLY_BOUND: dict[int, str | None] = {}


def is_available():
    """Return True if keybindings are available."""
    if Keybinder is None:
        return False

    try:
        return Keybinder.supported()
    except AttributeError:
        return True


def get_all_bound_keys() -> list[str]:
    """Return all currently bounded keys."""
    return list(filter(None, _CURRENTLY_BOUND.values()))


def get_current_event_time() -> int:
    "Return current event time as given by keybinder"
    if Keybinder is None:
        return 0

    return int(Keybinder.get_current_event_time())


def _register_bound_key(keystr: str | None, target: int) -> None:
    _CURRENTLY_BOUND[target] = keystr


def _get_currently_bound_key(
    target: KeybindingTarget = KEYBINDING_TARGET_DEFAULT,
) -> str | None:
    return _CURRENTLY_BOUND.get(target)


def bind_key(
    keystr: str | None,
    keybinding_target: KeybindingTarget = KEYBINDING_TARGET_DEFAULT,
) -> bool:
    """Bind @keystr, unbinding any previous key for @keybinding_target.
    If @keystr is a false value, any previous key will be unbound."""
    if Keybinder is None:
        return False

    if not _is_sane_keybinding(keystr):
        pretty.print_error(__name__, "Refusing to bind key", repr(keystr))
        return False

    succ = True
    if keystr:

        def callback(keystr: str) -> bool:
            get_keybound_object().keybinding(keybinding_target)
            return False

        try:
            succ = Keybinder.bind(keystr, callback)
            pretty.print_debug(__name__, "binding", repr(keystr))
            get_keybound_object().emit_bound_key_changed(keystr, True)
        except KeyError as exc:
            pretty.print_error(__name__, exc)
            succ = False

    if succ:
        old_keystr = _get_currently_bound_key(keybinding_target)
        if old_keystr and old_keystr != keystr:
            Keybinder.unbind(old_keystr)
            pretty.print_debug(__name__, "unbinding", repr(old_keystr))
            get_keybound_object().emit_bound_key_changed(old_keystr, False)

        _register_bound_key(keystr, keybinding_target)

    return succ


def _is_sane_keybinding(keystr: str | None) -> bool:
    """Refuse keys that we absolutely do not want to bind"""
    if keystr is None:
        return True

    if len(keystr) == 1 and keystr.isalnum():
        return False

    return keystr not in ("Return", "space", "BackSpace", "Escape")



"""
Acquired from https://github.com/quicksilver/Quicksilver
Which says

'''
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this software except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.

'''

Apache License 2-0 is GPLv3+ compatible --ulrik


"""

__kupfer_name__ = _("Quicksilver Icons")
__description__ = ""
__version__ = ""
__author__ = "Quicksilver Authors. License: Apache 2.0"

import typing as ty

if ty.TYPE_CHECKING:
    from gettext import gettext as _  # noqa


def initialize_plugin(name):
    pass



"""
Tracker plugins are versioned by the D-Bus API version
This is version works with Tracker1.
"""

__kupfer_name__ = _("Tracker")
__kupfer_sources__ = ()
__kupfer_text_sources__ = ("TrackerFulltext",)
__kupfer_contents__ = ("TrackerQuerySource",)
__kupfer_actions__ = ("TrackerSearch", "TrackerSearchHere")
__description__ = _("Tracker desktop search integration")
__version__ = "2017.2"
__author__ = "US"

import os
import typing as ty
from contextlib import suppress
from xml.etree.ElementTree import ElementTree

import dbus
from gi.repository import Gio

from kupfer import launch, plugin_support
from kupfer.obj import (
    Action,
    FileLeaf,
    OperationError,
    Source,
    TextLeaf,
    TextSource,
)
from kupfer.support import pretty

if ty.TYPE_CHECKING:
    from gettext import gettext as _

plugin_support.check_dbus_connection()

# SERVICE_NAME = "org.freedesktop.Tracker"
# SEARCH_OBJECT_PATH = "/org/freedesktop/Tracker/Search"
# SEARCH_INTERFACE = "org.freedesktop.Tracker.Search"

_SERVICE1_NAME = "org.freedesktop.Tracker1"
_SEARCH_OBJECT1_PATH = "/org/freedesktop/Tracker1/Resources"
_SEARCH1_INTERFACE = "org.freedesktop.Tracker1.Resources"

TRACKER_GUI_SEARCH = "tracker-needle"


class TrackerSearch(Action):
    def __init__(self):
        Action.__init__(self, _("Search in Tracker"))

    def activate(self, leaf, iobj=None, ctx=None):
        try:
            launch.spawn_async_raise([TRACKER_GUI_SEARCH, leaf.object])
        except launch.SpawnError as exc:
            raise OperationError(exc) from exc

    def get_description(self):
        return _("Open Tracker Search Tool and search for this term")

    def get_icon_name(self):
        return "system-search"

    def item_types(self):
        yield TextLeaf


class TrackerSearchHere(Action):
    action_accelerator = "t"

    def __init__(self):
        super().__init__(_("Get Tracker Results..."))

    def wants_context(self):
        return True

    def activate(self, leaf, iobj=None, ctx=None):
        assert ctx
        query = leaf.object

        def error(exc):
            ctx.register_late_error(exc)

        def reply(results):
            ret = []
            new_file = Gio.File.new_for_uri
            for result in results:
                with suppress(
                    Exception
                ):  # This very vague exception is from getpath
                    ret.append(FileLeaf(new_file(result[0]).get_path()))

            ctx.register_late_result(
                TrackerQuerySource(query, search_results=ret)
            )

        for _ignore in get_tracker_filequery(
            query,
            max_items=500,
            operation_err=True,
            reply_handler=reply,
            error_handler=error,
        ):
            pass

    def get_description(self):
        return _("Show Tracker results for query")

    def get_icon_name(self):
        return "tracker"

    def item_types(self):
        yield TextLeaf


def sparql_escape(ustr):
    """Escape unicode string @ustr for insertion into a SPARQL query

    Implemented to behave like tracker_sparql_escape in libtracker-client
    """
    sparql_escape_table = {
        ord("\t"): r"\t",
        ord("\n"): r"\n",
        ord("\r"): r"\r",
        ord("\b"): r"\b",
        ord("\f"): r"\f",
        ord('"'): r"\"",
        ord("\\"): "\\\\",
        # Extra rule: Can't have ?
        ord("?"): "",
    }
    return ustr.translate(sparql_escape_table)


ORDER_BY = {
    "rank": "ORDER BY DESC (fts:rank(?s))",
    "recent": "ORDER BY DESC (nfo:fileLastModified(?s))",
}


def get_file_results_sparql(
    searchobj, query, max_items=50, order_by="rank", location=None, **kwargs
):
    clean_query = sparql_escape(query)

    if location:
        location_filter = (
            "FILTER(tracker:uri-is-descendant "
            f'("{sparql_escape(location)}", nie:url (?s)))'
        )
    else:
        location_filter = ""

    sql = """SELECT tracker:coalesce (nie:url (?s), ?s)
              WHERE {
                ?s fts:match "%(query)s" .  ?s tracker:available true .
                %(location_filter)s
              }
              %(order_by)s
              OFFSET 0 LIMIT %(limit)d""" % {
        "query": clean_query,
        "location_filter": location_filter,
        "order_by": ORDER_BY[order_by],
        "limit": int(max_items),
    }

    pretty.print_debug(__name__, sql)
    results = searchobj.SparqlQuery(sql, **kwargs)
    if results is None:
        return

    new_file = Gio.File.new_for_uri
    for result in results:
        with suppress(Exception):  # This very vague exception is from getpath
            yield FileLeaf(new_file(result[0]).get_path())


_USE_VERSION = "0.8"
_VERSIONS = {
    "0.8": (_SERVICE1_NAME, _SEARCH_OBJECT1_PATH, _SEARCH1_INTERFACE),
}

version_query = {
    "0.8": get_file_results_sparql,
}


def get_searchobject(sname, opath, sinface, operation_err=False):
    bus = dbus.SessionBus()
    searchobj = None
    try:
        tobj = bus.get_object(sname, opath)
        searchobj = dbus.Interface(tobj, sinface)
    except dbus.DBusException as exc:
        if operation_err:
            raise OperationError(exc) from exc

        pretty.print_debug(__name__, exc)

    return searchobj


def get_tracker_filequery(query, operation_err=False, **kwargs):
    searchobj = get_searchobject(
        *_VERSIONS[_USE_VERSION], operation_err=operation_err
    )
    if searchobj is None:
        pretty.print_error(__name__, "Could not connect to Tracker")
        return ()

    queryfunc = version_query[_USE_VERSION]
    return queryfunc(searchobj, query, **kwargs)


class TrackerQuerySource(Source):
    def __init__(self, query, search_results=None, **search_args):
        Source.__init__(self, name=_('Tracker Search for "%s"') % query)
        self.query = query
        self.search_args = search_args
        self.search_results = None

    def repr_key(self):
        return self.query

    def get_items(self):
        if self.search_results:
            return self.search_results

        return get_tracker_filequery(self.query, **self.search_args)

    def provides(self):
        yield FileLeaf

    def get_description(self):
        return _('Results for "%s"') % self.query

    def get_icon_name(self):
        return "tracker"

    @classmethod
    def decorates_type(cls):
        return FileLeaf

    @classmethod
    def decorate_item(cls, leaf):
        # FIXME: Very simplified .savedSearch parsing, so far we only support
        # the query, without additional filtering. The simplest form of
        # .savedSearch file is saved by nautilus as following:
        # <query version="1.0">
        #   <text>QUERY GOES HERE</text>
        # </query>

        if not leaf.object.endswith(".savedSearch"):
            return None

        try:
            etree = ElementTree(file=leaf.object)
            text_elem = etree.getroot().find("text")
            if not text_elem:
                return None

            query = text_elem.text
            if not query:
                return None

            location_tag = etree.getroot().find("location")
            location = location_tag.text if location_tag is not None else None
            return cls(query, location=location_uri(location))

        except Exception:
            pretty.print_exc(__name__)
            return None


def location_uri(location):
    if location is None:
        return None

    if not os.path.isabs(location):
        location = os.path.expanduser("~/" + location)

    return Gio.File.new_for_path(location).get_uri()


# FIXME: Port tracker tag sources and actions
# to the new, much more powerful sparql + dbus API
# (using tracker-tag as in 0.6 is a plain hack and a dead end)


class TrackerFulltext(TextSource):
    def __init__(self):
        TextSource.__init__(self, name=_("Tracker Full Text Search"))

    def get_description(self):
        return _("Use '?' prefix to get full text results")

    def get_text_items(self, text):
        if text.startswith("?"):
            rank = "rank"
            if text.startswith("?~"):
                rank = "recent"

            query = text.lstrip("? ~")
            if len(query) > 2 and not has_parsing_error(query):  # noqa: PLR2004
                yield from TrackerQuerySource(
                    query, order_by=rank, max_items=50
                ).get_items()

    def provides(self):
        yield FileLeaf

    def get_rank(self):
        return 80


def has_parsing_error(query):
    "Check common parsing errors"
    words = query.split()
    # Unfinshed "" and OR, AND without following won't parse
    if words and words[-1] in ("OR", "AND"):
        return True

    return query.count('"') % 2 != 0



from __future__ import annotations

import os
import typing as ty
from contextlib import suppress
from dataclasses import dataclass
from pathlib import Path

import xdg.DesktopEntry
import xdg.Exceptions
from gi.repository import Gdk, Gio, GLib, Gtk

from kupfer.core import settings
from kupfer.support import desktop_parse, pretty
from kupfer.support import itertools as kitertools

__all__ = ["launch_app_info", "spawn_app", "spawn_app_id"]

STARTUP_ENV = "DESKTOP_STARTUP_ID"

# TODO: Broadcast Gio's launched message on dbus
# NOTE: GDK's startup notification things that we use
#       are really only sending xmessages. (roughly).


class SpawnError(Exception):
    """Error starting application"""


class ResourceLookupError(Exception):
    """Unable to find resource"""


class ResourceReadError(Exception):
    """Unable to open resource"""


def _find_desktop_file(desk_id: str) -> str:
    """Find file for @desk_id or raise ResourceLookupError

    Desktop files are found by appending /applications/ to
    $XDG_DATA_DIRS, but if they are located in subdirs of that,
    then additional 'subdirectory-' prefixes are used.
    """
    if not desk_id:
        raise ResourceLookupError("Empty id")

    try:
        return next(xdg.BaseDirectory.load_data_paths("applications", desk_id))  # type: ignore
    except StopIteration:
        ## it was not found as an immediate child of the data paths,
        ## so we split by the hyphens and search deeper
        file_id = desk_id
        directories = ["applications"]

        def lookup(path: list[str]) -> str | None:
            """Return location for @path if exists, else none"""
            return next(xdg.BaseDirectory.load_data_paths(*path), None)

        def get_dir_id_depth(desk_id: str, depth: int) -> tuple[str, str]:
            "split 'hyph-example-id' at the nth hyphen"
            parts = desk_id.split("-", depth)
            return "-".join(parts[:depth]), "-".join(parts[depth:])

        while True:
            ## try the first parts of the id to see if it matches a directory
            for x in range(1, 4):
                dirname, rest_id = get_dir_id_depth(file_id, x)
                if rest_id and lookup([*directories, dirname]):
                    file_id = rest_id
                    directories.append(dirname)
                    break
            else:
                ## we did not reach break
                break

            if desktop_file_path := lookup([*directories, file_id]):
                return desktop_file_path

    raise ResourceLookupError(f"Cannot locate '{desk_id}'")


def _read_desktop_info(desktop_file: str) -> dict[str, str | bool]:
    """Get the keys StartupNotify, Terminal, Exec, Path, Icon.
    Return dict with bool and unicode values."""
    # Return values in unicode
    try:
        entry = xdg.DesktopEntry.DesktopEntry(desktop_file)
    except xdg.Exceptions.Error as exc:
        raise ResourceReadError from exc

    if not entry.getExec():
        raise ResourceReadError("Invalid data: empty Exec key")

    return {
        "Terminal": entry.getTerminal(),
        "StartupNotify": entry.getStartupNotify(),
        "Exec": entry.getExec(),
        "Path": entry.getPath(),
        "Icon": entry.getIcon(),
        "Name": entry.getName(),
    }


def _app_to_desktop_info(app_info: Gio.AppInfo) -> dict[str, str | bool]:
    return {
        "Exec": app_info.get_commandline() or "",
        "Name": app_info.get_name(),
        "Icon": "",
        "Path": "",
        "Terminal": False,
        "StartupNotify": False,
    }


def _get_file_path(gfile: Gio.File) -> str:
    if gfile:
        return ty.cast("str", gfile.get_path() or gfile.get_uri())

    return ""


@dataclass
class _Flags:
    did_see_small_f: bool = False
    did_see_large_f: bool = False


# pylint: disable=too-many-locals
def _replace_format_specs(
    argv: list[str],
    location: str,
    desktop_info: dict[str, ty.Any],
    gfilelist: list[Gio.File],
) -> tuple[bool, bool, list[str]]:
    """
    http://standards.freedesktop.org/desktop-entry-spec/latest/ar01s06.html

    Replace format specifiers

    %% literal %
    %f file
    %F list of files
    %u URL
    %U list of URLs
    %i --icon <Icon key>
    %c Translated name
    %k location of .desktop file

    deprecated are removed:
    %d %D %n %N %v %m

    apart from those, all other.. stay and are ignored
    Like other implementations, we do actually insert
    a local path for %u and %U if it exists.

    Return (supports_single, added_at_end, argv)

    supports_single: Launcher only supports a single file
                     caller has to re-call for each file
    added_at_end:    No format found for the file, it was added
                     at the end
    """
    supports_single_file = False
    files_added_at_end = False
    flags = _Flags()
    fileiter = iter(gfilelist)

    def get_next_file_path() -> str:
        with suppress(StopIteration):
            file = next(fileiter)
            return _get_file_path(file)

        return ""

    # pylint: disable=too-many-return-statements
    def replace_single_code(key: str) -> str | None:  # noqa: PLR0911
        "Handle all embedded format codes, including those to be removed"
        if key in ("%d", "%D", "%n", "%N", "%v", "%m"):  # deprecated keys
            return ""

        if key == "%%":
            return "%"

        if key in ("%f", "%u"):
            if flags.did_see_large_f or flags.did_see_small_f:
                pretty.print_debug(
                    __name__, "Warning, multiple file format specs!"
                )
                return ""

            flags.did_see_small_f = True
            return get_next_file_path()

        if key == "%c":
            return desktop_info["Name"] or location

        if key == "%k":
            return location

        return None

    def replace_array_format(elem: str) -> tuple[bool, str | list[str]]:
        """Handle array format codes -- only recognized as single arguments

        Return  flag, arglist
        where flag is true if something was replaced
        """
        if elem in ("%U", "%F"):
            if flags.did_see_large_f or flags.did_see_small_f:
                pretty.print_debug(
                    __name__, "Warning, multiple file format specs!"
                )
                return True, []

            flags.did_see_large_f = True
            return True, list(filter(bool, map(_get_file_path, gfilelist)))

        if elem == "%i":
            if desktop_info["Icon"]:
                return True, ["--icon", desktop_info["Icon"]]

            return True, []

        return False, elem

    new_argv: list[str] = []
    for x in argv:
        if not x:
            # the arg is an empty string, we don't need extra processing
            new_argv.append(x)
            continue

        succ, newargs = replace_array_format(x)
        if succ:
            new_argv.extend(newargs)
        elif arg := kitertools.two_part_mapper(x, replace_single_code):
            # Handle embedded format codes
            new_argv.append(arg)

    if len(gfilelist) > 1 and not flags.did_see_large_f:
        supports_single_file = True

    if not flags.did_see_small_f and not flags.did_see_large_f and gfilelist:
        files_added_at_end = True
        new_argv.append(get_next_file_path())

    return supports_single_file, files_added_at_end, new_argv


def _file_for_app_info(app_info: Gio.AppInfo) -> str | None:
    try:
        return _find_desktop_file(app_info.get_id())
    except ResourceLookupError:
        pretty.print_exc(__name__)

    return None


def _info_for_desktop_file(
    desktop_file: str | None,
) -> dict[str, ty.Any] | None:
    if desktop_file:
        try:
            return _read_desktop_info(desktop_file)
        except ResourceReadError:
            pretty.print_exc(__name__)

    return None


# pylint: disable=too-few-public-methods
@ty.runtime_checkable
class LaunchCallback(ty.Protocol):
    def __call__(
        self,
        argv: list[str],
        pid: int,
        notify_id: int | None,
        filelist: list[str],
        timestamp: int | None,
        /,
    ) -> None: ...


# pylint: disable=too-many-locals
def launch_app_info(  # noqa:PLR0912
    app_info: Gio.AppInfo,
    gfiles: ty.Iterable[Gio.File] | None = None,
    in_terminal: bool | None = None,
    timestamp: int | None = None,
    desktop_file: str | None = None,
    launch_cb: LaunchCallback | None = None,
    screen: Gdk.Screen | None = None,
    work_dir: str | None = None,
) -> bool:
    """Launch @app_info, opening @gfiles.

    @in_terminal: override Terminal flag
    @timestamp: override timestamp
    @desktop_file: specify location of desktop file
    @launch_cb: Called once per launched process, like ``spawn_app``
    @work_dir: overwrite work dir of application. By default it's
    "Path" parameter in desktop file.

    Will pass on exceptions from spawn_app
    """
    gfiles = list(gfiles or [])
    desktop_file = desktop_file or _file_for_app_info(app_info)
    desktop_info = _info_for_desktop_file(desktop_file)
    if not desktop_file or not desktop_info:
        # Allow in-memory app_info creations (without id or desktop file)
        desktop_file = ""
        desktop_info = _app_to_desktop_info(app_info)
        # in this case, the command line is already primarily escaped
        desktop_exe = desktop_info["Exec"]
        assert desktop_exe and isinstance(desktop_exe, str)
        argv = desktop_parse.parse_argv(desktop_exe)
    else:
        # In the normal case, we must first escape one round
        argv = desktop_parse.parse_unesc_argv(desktop_info["Exec"])

    assert argv and argv[0]

    # Now Resolve the %f etc format codes
    multiple_needed, _missing_format, launch_argv = _replace_format_specs(
        argv, desktop_file, desktop_info, gfiles
    )

    if not multiple_needed:
        # Launch 1 process
        launch_records = [(launch_argv, gfiles)]
    else:
        # Launch one process per file
        launch_records = [(launch_argv, [gfiles[0]])]
        for file in gfiles[1:]:
            *_ignore, launch_argv = _replace_format_specs(
                argv, desktop_file, desktop_info, [file]
            )
            launch_records.append((launch_argv, [file]))

    notify = bool(desktop_info["StartupNotify"])
    if not work_dir:
        work_dir = ty.cast("ty.Optional[str]", desktop_info["Path"])

    if in_terminal is None:
        in_terminal = bool(desktop_info["Terminal"])

    if in_terminal:
        term = settings.get_configured_terminal()
        if term:
            notify = notify or bool(term["startup_notify"])

    for argv, files in launch_records:
        if in_terminal and term:
            targv = list(term["argv"])
            if exearg := term["exearg"]:
                targv.append(exearg)

            argv = targv + argv  # noqa: PLW2901

        if not spawn_app(
            app_info,
            argv,
            files,
            work_dir,
            notify,
            timestamp=timestamp,
            launch_cb=launch_cb,
            screen=screen,
        ):
            return False

    return True


def spawn_app_id(
    app_id: str,
    argv: list[str],
    workdir: str | None = None,
    startup_notify: bool = True,
    screen: Gdk.Screen | None = None,
) -> bool:
    """Spawn @argv trying to notify it as if it is app_id."""
    try:
        app_info = _get_info_for_id(app_id)
    except (TypeError, RuntimeError):
        app_info = None
        startup_notify = False

    return bool(
        spawn_app(app_info, argv, [], workdir, startup_notify, screen=screen)
    )


def spawn_app(
    app_info: Gio.AppInfo | None,
    argv: list[str],
    filelist: list[ty.Any],
    workdir: str | None = None,
    startup_notify: bool = True,
    timestamp: int | None = None,
    launch_cb: LaunchCallback | None = None,
    screen: Gdk.Screen | None = None,
) -> int:
    """Spawn app.

    @argv: argument list including files
    @workdir: where to set workdir if not cwd
    @app_info: Used for startup notification, if @startup_notify is True
    @filelist: Used for startup notification
    @startup_notify: Use startup notification
    @timestamp: Event timestamp
    @launch_cb: Called if successful with
                (argv, pid, notify_id, filelist, timestamp)
    @screen: GdkScreen on which to put the application

    return pid if successful
    raise SpawnError on error
    """
    notify_id = None
    if startup_notify:
        display = Gdk.Display.get_default()
        ctx = display.get_app_launch_context()
        ctx.set_timestamp(timestamp or Gtk.get_current_event_time())
        if screen:
            ctx.set_screen(screen)
        # This not only returns the string ID but
        # it actually starts the startup notification!
        notify_id = ctx.get_startup_notify_id(app_info, filelist)
        child_env_add = {STARTUP_ENV: notify_id}
    else:
        child_env_add = {}

    if not workdir or not Path(workdir).exists():
        workdir = "."

    argv_ = list(_locale_encode_argv(argv))
    pid: int

    try:
        pid, *_ig = GLib.spawn_async(
            argv_,
            flags=GLib.SpawnFlags.SEARCH_PATH,
            working_directory=workdir,
            child_setup=child_setup,
            user_data=child_env_add,
        )
        pretty.print_debug(__name__, "Launched", argv_, notify_id, "pid:", pid)

    except GLib.GError as exc:
        pretty.print_error(__name__, "Error Launching ", argv_, str(exc))
        if notify_id:
            Gdk.notify_startup_complete_with_id(notify_id)

        raise SpawnError(exc.message) from exc  # pylint: disable=no-member

    if launch_cb:
        launch_cb(argv, pid, notify_id, filelist, timestamp)

    return pid


def child_setup(add_environ: dict[str, str]) -> None:
    """Called to setup the child process before exec()
    @add_environ is a dict for extra env variables
    """
    for key, val in add_environ.items():
        os.putenv(key, val)


def _locale_encode_argv(argv: list[str] | list[bytes]) -> ty.Iterator[str]:
    for x in argv:
        if isinstance(x, str):
            yield x
        else:
            yield x.decode("UTF-8", "replace")


def _get_info_for_id(app_id: str) -> Gio.DesktopAppInfo:
    return Gio.DesktopAppInfo.new(app_id)


if __name__ == "__main__":
    while True:
        id_ = input("Give me an App ID > ")
        launch_app_info(_get_info_for_id(id_ + ".desktop"), [])
        # launch_app_info(Gio.AppInfo("gvim"), [Gio.File(".")])



"""
This module is a part of the program Kupfer, see the main program file for
more information.
"""

from __future__ import annotations

import sys
import typing as ty
import weakref

if ty.TYPE_CHECKING:
    from gi.repository import GObject

__all__ = (
    "DbusWeakCallback",
    "GobjectWeakCallback",
    "WeakCallback",
    "gobject_connect_weakly",
)

MCallback = ty.Callable[..., None]


class WeakCallback:
    """A Weak Callback object that will keep a reference to
    the connecting object with weakref semantics.

    This allows object A to pass a callback method to object S,
    without object S keeping A alive.
    """

    def __init__(self, mcallback: MCallback) -> None:
        """Create a new Weak Callback calling the method @mcallback"""
        # pylint: disable=no-member
        obj = mcallback.__self__  # type: ignore
        # pylint: disable=no-member
        attr = mcallback.__func__.__name__  # type: ignore
        self.wref = weakref.ref(obj, self.object_deleted)
        self.callback_attr = attr
        self.token = None

    def __call__(self, *args: ty.Any, **kwargs: ty.Any) -> None:
        if obj := self.wref():
            attr = getattr(obj, self.callback_attr)
            attr(*args, **kwargs)
        else:
            self.default_callback(*args, **kwargs)

    def default_callback(self, *args: ty.Any, **kwargs: ty.Any) -> None:
        """Called instead of callback when expired"""

    def object_deleted(self, wref: weakref.ReferenceType[ty.Any]) -> None:
        """Called when callback expires"""


class DbusWeakCallback(WeakCallback):
    """
    Will use @token if set as follows:
        token.remove()
    """

    def object_deleted(self, wref: weakref.ReferenceType[ty.Any]) -> None:
        # App is shutting down
        try:
            if sys.is_finalizing():
                return
        except AttributeError:
            if sys is None:
                return

        if self.token:
            self.token.remove()
            self.token = None


def dbus_signal_connect_weakly(
    bus: ty.Any, signal: str, mcallback: MCallback, **kwargs: ty.Any
) -> None:
    """
    Connect method @mcallback to dbus signal using a weak callback

    Connect to @signal on @bus, passing on all keyword arguments
    """
    weak_cb = DbusWeakCallback(mcallback)
    weak_cb.token = bus.add_signal_receiver(weak_cb, signal, **kwargs)


class GobjectWeakCallback(WeakCallback):
    """
    Will use @token if set as follows:
        sender.disconnect(token)
    """

    __senders: ty.ClassVar[dict[ty.Any, GObject]] = {}

    def object_deleted(self, wref: weakref.ReferenceType[GObject]) -> None:
        # App is shutting down
        try:
            if sys.is_finalizing():
                return
        except AttributeError:
            if sys is None:
                return

        sender = self.__senders.pop(self.token, None)
        if sender is not None:
            sender.disconnect(self.token)

    @classmethod
    def _connect(
        cls,
        sender: GObject,
        signal: str,
        mcallback: MCallback,
        *user_args: ty.Any,
    ) -> None:
        # We save references to the sender in a class variable,
        # this is the only way to have it accessible when obj expires.
        wcb = cls(mcallback)
        wcb.token = sender.connect(signal, wcb, *user_args)
        cls.__senders[wcb.token] = sender


def gobject_connect_weakly(
    sender: GObject, signal: str, mcallback: MCallback, *user_args: ty.Any
) -> None:
    """Connect weakly to GObject @sender's @signal,
    with a callback method @mcallback

    >>> import gi
    >>> gi.require_version("Gtk", "3.0")
    >>> from gi.repository import Gtk
    >>> btn = Gtk.Button.new()
    >>> class Handler (object):
    ...   def handle(self): pass
    ...   def __del__(self): print("deleted")
    ...
    >>> h = Handler()
    >>> gobject_connect_weakly(btn, "clicked", h.handle)
    >>> del h
    deleted
    >>>
    """
    # pylint: disable=protected-access
    GobjectWeakCallback._connect(  # noqa: SLF001
        sender, signal, mcallback, *user_args
    )


if __name__ == "__main__":
    import doctest

    doctest.testmod()



"""
Thunderbird  plugin.

Changes:
    2012-03-15: Karol Będkowski
        + activate_multiple for new mail action
"""

__kupfer_name__ = _("Thunderbird")
__kupfer_sources__ = ("ContactsSource",)
__kupfer_actions__ = (
    "NewMailAction",
    "AttachToNewMail",
    "NewMailWithBody",
)
__description__ = _("Thunderbird/Icedove Contacts and Actions")
__version__ = "2017.2"
__author__ = "Karol Będkowski <karol.bedkowski@gmail.com>, US"

import typing as ty

from kupfer import icons, launch, plugin_support
from kupfer.obj import Action, FileLeaf, RunnableLeaf, TextLeaf, UrlLeaf
from kupfer.obj.apps import AppLeafContentMixin
from kupfer.obj.contacts import ContactLeaf, EmailContact, email_from_leaf
from kupfer.obj.grouping import ToplevelGroupingSource
from kupfer.obj.helplib import FilesystemWatchMixin
from kupfer.plugin import thunderbird_support as support

if ty.TYPE_CHECKING:
    from gettext import gettext as _


plugin_support.check_any_command_available("thunderbird", "icedove")


class ComposeMail(RunnableLeaf):
    """Create new mail without recipient"""

    def __init__(self):
        RunnableLeaf.__init__(self, name=_("Compose New Email"))

    def run(self, ctx=None):
        if not launch.spawn_async(["thunderbird", "--compose"]):
            launch.spawn_async(["icedove", "--compose"])

    def get_description(self):
        return _("Compose a new message in Thunderbird")

    def get_icon_name(self):
        return "mail-message-new"


class NewMailAction(Action):
    """Create new mail to selected leaf (Contact or TextLeaf)"""

    def __init__(self):
        Action.__init__(self, _("Compose Email To"))

    def activate(self, leaf, iobj=None, ctx=None):
        self.activate_multiple((leaf,))

    def activate_multiple(self, objects):
        recipients = ",".join(
            filter(None, (email_from_leaf(L) for L in objects))
        )
        if not recipients:
            return

        if not launch.spawn_async(["thunderbird", f"mailto:{recipients}"]):
            launch.spawn_async(["icedove", f"mailto:{recipients}"])

    def get_icon_name(self):
        return "mail-message-new"

    def item_types(self):
        yield ContactLeaf
        # we can enter email
        yield TextLeaf
        yield UrlLeaf

    def valid_for_item(self, leaf):
        return bool(email_from_leaf(leaf))


class AttachToNewMail(Action):
    def __init__(self):
        super().__init__(_("Attach in Email To..."))

    def activate(self, leaf, iobj=None, ctx=None):
        assert iobj
        self.activate_multiple((leaf,), (iobj,))

    def activate_multiple(self, objects, iobjects):
        attachments = ",".join(L.object for L in objects)
        recipients = ",".join(
            filter(None, (email_from_leaf(L) for L in iobjects))
        )
        if not recipients:
            return

        args = ["-compose", f"to='{recipients}',attachment='{attachments}'"]
        if not launch.spawn_async(["thunderbird", *args]):
            launch.spawn_async(["icedove", *args])

    def get_icon_name(self):
        return "mail-message-new"

    def item_types(self):
        yield FileLeaf

    def valid_for_item(self, leaf):
        return not leaf.is_dir()

    def requires_object(self):
        return True

    def object_types(self):
        yield ContactLeaf
        # we can enter email
        yield TextLeaf
        yield UrlLeaf

    def valid_object(self, iobj, for_item=None):
        return bool(email_from_leaf(iobj))


class NewMailWithBody(Action):
    def __init__(self):
        super().__init__(_("Compose Email With"))

    def activate(self, leaf, iobj=None, ctx=None):
        args = ["-compose", f"body='{leaf.object}'"]
        if not launch.spawn_async(["thunderbird", *args]):
            launch.spawn_async(["icedove", *args])

    def get_icon_name(self):
        return "mail-message-new"

    def item_types(self):
        yield TextLeaf

    def get_description(self):
        return _("Compose a new message using the text as body")


class ContactsSource(
    AppLeafContentMixin, ToplevelGroupingSource, FilesystemWatchMixin
):
    appleaf_content_id = ("thunderbird", "icedove")
    source_scan_interval: int = 3600

    def __init__(self, name=_("Thunderbird Address Book")):
        ToplevelGroupingSource.__init__(self, name, "Contacts")
        self._version = 2
        self.monitor_token = None

    def initialize(self):
        ToplevelGroupingSource.initialize(self)
        if abook_dirs := list(support.get_addressbook_dirs()):
            self.monitor_token = self.monitor_directories(*abook_dirs)

    def monitor_include_file(self, gfile):
        return gfile and (
            gfile.get_basename().endswith(".mab")
            or gfile.get_basename() == "localstore.rdf"
        )

    def mark_for_update(self, postpone=False):
        super().mark_for_update(postpone=True)

    def get_items(self):
        for name, email in support.get_contacts():
            yield EmailContact(email, name)

        yield ComposeMail()

    def should_sort_lexically(self):
        return True

    def get_description(self):
        return _("Contacts from Thunderbird Address Book")

    def get_gicon(self):
        return icons.get_gicon_with_fallbacks(None, ("thunderbird", "icedove"))

    def provides(self):
        yield ContactLeaf
        yield RunnableLeaf



import fnmatch
import io
import pickle
import sys
import typing as ty

__all__ = ("BasicUnpickler", "ConservativeUnpickler")


# pylint: disable=too-few-public-methods
class UniversalSet:
    def __contains__(self, item: ty.Any) -> bool:
        return True


class ConservativeUnpickler(pickle.Unpickler):
    """An Unpickler that refuses to import new modules

    >>> import pickle

    >>> import kupfer.objects
    >>> ConservativeUnpickler.loads(pickle.dumps(kupfer.objects.FileLeaf("A")))
    <builtin.FileLeaf A>

    >>> ConservativeUnpickler.loads(pickle.dumps(eval))
    Traceback (most recent call last):
        ...
    UnpicklingError: Refusing unsafe __builtin__.eval

    >>> import sys
    >>> import kupfer.obj.base
    >>> pdata = pickle.dumps(kupfer.obj.base.Leaf(1, "A"))
    >>> del sys.modules["kupfer.obj.base"]
    >>> ConservativeUnpickler.loads(pdata)
    Traceback (most recent call last):
        ...
    UnpicklingError: Refusing to load module kupfer.obj.base
    """

    safe_modules: ty.ClassVar[dict[str, ty.Any]] = {
        "builtins": {"set", "sum", "object"},
        "copy_reg": {"_reconstructor"},
        "kupfer.*": UniversalSet(),
    }

    @classmethod
    def is_safe_symbol(cls, module: str, name: str) -> bool:
        for pattern, modules in cls.safe_modules.items():
            if fnmatch.fnmatchcase(module, pattern):
                return name in modules

        return False

    def find_class(self, module: str, name: str) -> ty.Any:
        if module not in sys.modules:
            raise pickle.UnpicklingError(f"Refusing to load module {module}")

        if not self.is_safe_symbol(module, name):
            raise pickle.UnpicklingError(f"Refusing unsafe {module}.{name}")

        return pickle.Unpickler.find_class(self, module, name)

    @classmethod
    def loads(cls, pickledata: bytes) -> ty.Any:
        unpickler = cls(io.BytesIO(pickledata))
        return unpickler.load()


class BasicUnpickler(ConservativeUnpickler):
    """An Unpickler that can only unpickle persistend ids and select builtins

    >>> import pickle

    >>> import kupfer.objects
    >>> BasicUnpickler.loads(pickle.dumps(kupfer.objects.FileLeaf("A")))
    Traceback (most recent call last):
        ...
    UnpicklingError: Refusing unsafe kupfer.obj.objects.FileLeaf
    """

    safe_modules: ty.ClassVar[dict[str, set[str]]] = {
        "__builtin__": {"object"},
        "copy_reg": {"_reconstructor"},
        "kupfer.puid": {"SerializedObject"},
    }


if __name__ == "__main__":
    import doctest

    doctest.testmod()
